/*
 * file bomb.c - bombs and explosions
 *
 * $Id: bomb.c,v 1.29 2001/07/01 16:01:23 xblast Exp $
 *
 * Program XBLAST 
 * (C) 1993-1999 by Oliver Vogel (e-mail: m.vogel@ndh.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will be entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "bomb.h"

#include "atom.h"
#include "func.h"
#include "geom.h"
#include "info.h"
#include "map.h"
#include "snd.h"
#include "random.h"

/*
 * macros
 */
#define BOMB_ERROR_PROB 16
#define BOMB_DELAY 5

#define BOMB_VX        (2*BASE_X)
#define BOMB_VY        (2*BASE_Y)

#define BOMB_STUN_X    (4*BASE_X)
#define BOMB_STUN_Y    (4*BASE_Y)

/* fuse times */
#define BOMB_TIME   64
#define SHORT_FUSE  32
#define LONG_FUSE  128

/* haunt factors */
#define HAUNT_NONE 0
#define HAUNT_FAST 10
#define HAUNT_SLOW 35

/* nasty walls */
#define NASTY_INC GAME_TIME/12

/*
 * local types
 */

/* burn out flags */
typedef enum {
  BO_RIGHT = (1<<0),
  BO_LEFT  = (1<<1),
  BO_UP    = (1<<2),
  BO_DOWN  = (1<<3),
  BO_TOTAL = (1<<4)-1
} BMBurnOut;

/* bomb and explosion */
typedef struct explosion {
  BMPlayer   	   *player;      /* from which player */
  int        	    range;       /* blast range */
  int        	    x, y;        /* grid position */
  int        	    dx, dy;      /* small deviations from x and y (in pixel) */
  BMDirection	    dir;         /* Direction */
  XBBool       	    malfunction; /* flag for malfunction illness */
  int        	    count;       /* time counter */
  int        	    blink;       /* when is bomb due to blick */
  BMBombType 	    type;        /* what type of bomb do we have */
  int        	    typeExtr;    /* extra information for type */
  BMBurnOut  	    burnOut;     /* has bomb already burned out */
  Sprite     	   *sprite;      /* sprite showing bomb */
  int        	    anime;       /* current animation phase */
  int        	    nextAnime;   /* at which count is next animation due */
  XBBool       	    isMorphed;   /* player has morphed into this bomb */
  struct explosion *next;        /* pointer to next bomb in list */
} Explosion;

/* pointer to function with type specifix behaviour */
typedef XBBool (*XBSpecialBombFunc) (Explosion *);

/* pointer to function when bomb hits player, wall or other bomb */
typedef void (*XBBombClickFunc) (Explosion *);

/*
 * local prototypes
 */

/* "constructor" for a bomb */
static XBBool NewExplosion (BMPlayer *player, int x, int y, int range, XBBool remoteControlled, XBBool malfunction, 
			    BMBombType type, int typeExtr, BMDirection initialdir);
/* special bomb functions */
static XBBool SpecialBombNormal (Explosion *ptr);
static XBBool SpecialBombNapalm (Explosion *ptr);
static XBBool SpecialBombFirecracker (Explosion *ptr);
static XBBool SpecialBombConstruction (Explosion *ptr);
static XBBool SpecialBombThreebombs (Explosion *ptr);
static XBBool SpecialBombGrenade (Explosion *ptr);
static XBBool SpecialBombTrianglebombs (Explosion *ptr);
static XBBool SpecialBombDestruction (Explosion *ptr);
static XBBool SpecialBombFungus (Explosion *ptr);
static XBBool SpecialBombRenovation (Explosion *ptr);
static XBBool SpecialBombPyro (Explosion *ptr);
/* bomb click functions */
static void BombClickNone (Explosion *bomb);
static void BombClickInitial (Explosion *bomb);
static void BombClickThru (Explosion *bomb);
static void BombClickSnooker (Explosion *bomb);
static void BombClickContact (Explosion *bomb);
static void BombClickClockwise (Explosion *bomb);
static void BombClickAnticlockwise (Explosion *bomb);
static void BombClickRandomdir (Explosion *bomb);
static void BombClickRebound (Explosion *bomb);

/*
 * local variables
 */

/* array to locate bomb/explosion on map */
static Explosion *bombMaze[MAZE_W][MAZE_H];
/* list with all bombs and explosions */
static int numExpl = 0 ;
static Explosion *explList;                
static Explosion *explEnd;
static int initialBombDir;
static int hauntFactor;
static int curBombTime;
static int defaultBMT, specialBMT, evilBMT;
static int nextNasty;
static int gentleNasty;
static int rangeNasty;
/* bomb click functions */
static XBBombClickFunc doBombClick;
static XBBombClickFunc doWallClick;
static XBBombClickFunc doPlayerClick;
/* lookup table for special bomb function */
static XBSpecialBombFunc doSpecialBombFunction[NUM_BMT] = {
  SpecialBombNormal,
  SpecialBombNapalm,
  SpecialBombNormal,
  SpecialBombNormal,
  SpecialBombFirecracker,
  SpecialBombFirecracker,
  SpecialBombConstruction,
  SpecialBombThreebombs,
  SpecialBombGrenade,
  SpecialBombTrianglebombs,
  SpecialBombDestruction,
  SpecialBombFungus,
  SpecialBombRenovation,
  SpecialBombPyro,
  SpecialBombPyro,
  SpecialBombNormal,
  SpecialBombNormal,
};
/* conversion table for bomb clicks */
static DBToData bombClickTable[] = {
  { "anticlockwise", (void *) BombClickAnticlockwise },
  { "clockwise",     (void *) BombClickClockwise },
  { "contact",       (void *) BombClickContact },
  { "initial",       (void *) BombClickInitial },
  { "randomdir",     (void *) BombClickRandomdir },
  { "rebound",       (void *) BombClickRebound },
  { "snooker",       (void *) BombClickSnooker },
  { "thru",          (void *) BombClickThru },
  { NULL,            (void *) BombClickNone },
};
/* conversion table for direction */
static DBToInt bombDirTable[] = {
  { "down",  GoDown },
  { "left",  GoLeft },
  { "right", GoRight },
  { "up",    GoUp },
  { NULL,    GoStop }
};
/* conversion table for fuse times */
static DBToInt fuseTimeTable[] = {
  { "long",  LONG_FUSE },
  { "short", SHORT_FUSE },
  { NULL,    BOMB_TIME }
};
/* conversion table for bomb types */
static DBToInt bombTypeTable[] = {
  { "blastnow",       BMTblastnow }, 
  { "close",          BMTclose }, 
  { "construction",   BMTconstruction }, 
  { "destruction",    BMTdestruction }, 
  { "firecracker",    BMTfirecracker }, 
  { "firecracker2",   BMTfirecracker2 }, 
  { "fungus",         BMTfungus }, 
  { "grenade",        BMTgrenade }, 
  { "napalm",         BMTnapalm }, 
  { "pyro",           BMTpyro }, 
  { "pyro2",          BMTpyro2 }, 
  { "random",         BMTrandom }, 
  { "renovation",     BMTrenovation }, 
  { "short",          BMTshort },
  { "threebombs",     BMTthreebombs }, 
  { "trianglebombs",  BMTtrianglebombs }, 
  { NULL,             BMTdefault },
};
/* bomb haunting */
static DBToInt hauntFactorTable[] = {
  { "fast", HAUNT_FAST },
  { "slow", HAUNT_SLOW },
  { NULL,   HAUNT_NONE },
};
/* bomb extra info */
static char *bombName[] = {
  NULL,
  "Napalm bomb",
  "Instant bomb",
  NULL,
  "Firecracker",
  NULL,
  "Construction bomb",
  "Three bomb",
  "Grenade",
  "Triangle bomb",
  "Destruction bomb",
  "Fungus bomb",
  "Renovation bomb",
  "Pyro bomb",
  NULL,
  "Random bomb",
  "Fast bomb",
};
/* bomb direction info */
static const char *bombDirInfo[MAX_DIR] = {
  NULL,
  "Bombs are going up",
  "Bombs are going left",
  "Bombs are falling down",
  "Bombs are going right",
  NULL,
};

/*
 * load bombs from level
 */
XBBool
ConfigLevelBombs (const DBSection *section)
{
  void *ptr;
  const char *s;

  assert (section != NULL);

  assert (NULL == explList);
  assert (NULL == explEnd);
  assert (0    == numExpl);
  /* clear list */
  ptr = NULL;
  /* clear lookup table */
  memset (bombMaze, 0, sizeof (bombMaze) );
  /* set bomb click behavior */
  if (DB_ConvertEntryData (section, atomBombClick, &ptr, bombClickTable) ) {
    doBombClick = (XBBombClickFunc) ptr;
  } else {
    doBombClick = BombClickNone;
  }
  if (DB_ConvertEntryData (section, atomWallClick, &ptr, bombClickTable) ) {
    doWallClick = (XBBombClickFunc) ptr;
  } else {
    doWallClick = BombClickNone;
  }
  if (DB_ConvertEntryData (section, atomPlayerClick, &ptr, bombClickTable) ) {
    doPlayerClick = (XBBombClickFunc) ptr;
  } else {
    doPlayerClick = BombClickNone;
  }
  /* general behavior */
  if (! DB_ConvertEntryInt (section, atomDirection, (int *) &initialBombDir, bombDirTable) ) {
    initialBombDir = GoStop;
  }
  if (! DB_ConvertEntryInt (section, atomFuseTime, &curBombTime, fuseTimeTable) ) {
    curBombTime = BOMB_TIME;
  }
  if (! DB_ConvertEntryInt (section, atomHaunt, &hauntFactor, hauntFactorTable) ) {
    hauntFactor = HAUNT_NONE;
  }
  /* bomb types */
  if (! DB_ConvertEntryInt (section, atomDefault, (int *) &defaultBMT, bombTypeTable) ) {
    defaultBMT = BMTnormal;
  }
  if (! DB_ConvertEntryInt (section, atomSpecial, (int *) &specialBMT, bombTypeTable) ) {
    specialBMT = BMTnormal;
  }
  if (! DB_ConvertEntryInt (section, atomHidden, (int *) &evilBMT, bombTypeTable) ) {
    evilBMT = BMTnormal;
  }
  /* nasty walls .. */
  if (DB_GetEntryInt (section, atomNastyGentle, &gentleNasty) &&
      DB_GetEntryInt (section, atomNastyRange, &rangeNasty) ) {
    /* ... activated */
    nextNasty    = NASTY_INC;
    gentleNasty *= GAME_TIME;
  }  else {
    /* ... deactivated */
    nextNasty   = GAME_TIME + NASTY_INC + 1;
    gentleNasty = 0;
    rangeNasty  = 0;
  }
  /* 
   * set level info text 
   */
  /* bomb types */
  if (NULL != bombName[defaultBMT]) {
    AddLevelInfo ("%s as default", bombName[defaultBMT]);
  }
  if (0 != NumSpecialBombs () && 
      NULL != bombName[specialBMT]) {
    AddLevelInfo ("%d %ss", NumSpecialBombs (), bombName[specialBMT]);
  }
  if (HasSpecialBombs () &&
      NULL != bombName[specialBMT]) {
    AddLevelInfo ("%s as an extra", bombName[specialBMT]);
  }
  if (NULL != bombName[evilBMT]) {
    AddLevelInfo ("Hidden %ss", bombName[evilBMT]);
  }
  /* fuse time */
  switch (curBombTime) {
  case LONG_FUSE:  AddLevelInfo ("All bombs are long fused");  break;
  case SHORT_FUSE: AddLevelInfo ("All bombs are short fused"); break;
  default:         break;
  }
  /* direction */
  if (NULL != (s = bombDirInfo[initialBombDir]) ) {
    AddLevelInfo (s);
  }
  /* haunting */
  switch (hauntFactor) {
  case HAUNT_SLOW: AddLevelInfo ("All bombs are haunted"); break;
  case HAUNT_FAST: AddLevelInfo ("All bombs are haunted (and dangerous)"); break;
  default:         break;
  }
  /* bomb click */
  if (doBombClick == BombClickRebound) {
    AddLevelInfo ("Bombs rebound from others");
  } else if (doBombClick == BombClickContact) {
    AddLevelInfo ("Bombs explode on contact with others");
  } else if (doBombClick == BombClickClockwise) {
    AddLevelInfo ("Bombs turn clockwise on hitting others");
  } else if (doBombClick == BombClickAnticlockwise) {
    AddLevelInfo ("Bombs turn anticlockwise on hitting others");
  } else if (doBombClick == BombClickRandomdir) {
    AddLevelInfo ("Bombs bounce off randomly from others");
  } else if (doBombClick == BombClickSnooker) {
    AddLevelInfo ("Bombs are snooker bombs");
  }
  /* wall click */
  if (doWallClick == BombClickRebound) {
    AddLevelInfo ("Bombs rebound off walls");
  } else if (doWallClick == BombClickContact) {
    AddLevelInfo ("Bombs explode on contact with walls");
  } else if (doWallClick == BombClickClockwise) {
    AddLevelInfo ("Bombs turn clockwise on hitting walls");
  } else if (doWallClick == BombClickAnticlockwise) {
    AddLevelInfo ("Bombs turn anticlockwise on hitting walls");
  } else if (doWallClick == BombClickRandomdir) {
    AddLevelInfo ("Bombs bounce off randomly of walls");
  }
  /* player click */
  if (doPlayerClick == BombClickThru) {
    AddLevelInfo ("Bombs stun players running through");
  } else if (doPlayerClick == BombClickContact) {
    AddLevelInfo ("Bombs explode on contact with players");
  } else if (doPlayerClick == BombClickRebound) {
    AddLevelInfo ("Bombs rebound off players");
  }  
  /* nasty walls */
  if (gentleNasty != 0) {
    AddLevelInfo ("The Walls launch bombs");
  }
  /* that's all */
  return XBTrue;
} /* ConfigBombs */

/*
 *
 */
static void 
OneExplAt (int x, int y, int ra, int ri, BMBurnOut *burnOut)
{
  int i;

  assert (NULL != burnOut);
  /* right */
  if (0 == (*burnOut & BO_RIGHT) ) { 
    for (i = 0; (i <= ra) && CheckMazeOpen(x+i, y) ; i ++) {
      if (i >= ri) {
	if (i != ri && i != ra) {
	  SetExplBlock (x+i, y, 0x1a);
	} else if (i != ri) {
	  SetExplBlock (x+i, y, 0x18);
	} else if (i == ra) {
	  SetExplBlock (x+i, y, 0x10);
	} else {
	  SetExplBlock (x+i, y, 0x12);
	}
      }
    }
    if (i < ri) {
      *burnOut |= BO_RIGHT;
    }
    if ( (i <= ra) && (CheckMazeExtra (x+i, y)) ) {
      SetMazeBlock (x+i, y, BTExtraOpen);
    }
  }
  /* left */
  if (0 == (*burnOut & BO_LEFT) ) { 
    for (i = 0; (i <= ra) && CheckMazeOpen (x-i, y); i ++) {
      if (i >= ri) {
	if (i != ri && i != ra) {
	  SetExplBlock (x-i, y, 0x1a);
	} else if (i != ri) {
	  SetExplBlock (x-i, y, 0x12);
	} else if (i == ra) {
	  SetExplBlock (x-i, y, 0x10);
	} else {
	  SetExplBlock (x-i, y, 0x18);
	}
      }
    }
    if (i < ri) {
      *burnOut |= BO_LEFT;
    }
    if ( (i <= ra ) && (CheckMazeExtra(x-i, y)) ) {
      SetMazeBlock (x-i, y, BTExtraOpen);
    }
  }
  /* up */
  if (0 == (*burnOut & BO_UP) ) { 
    for (i = 0; (i <= ra) && CheckMazeOpen (x, y-i); i ++) {
      if (i >= ri) {
	if (i != ri && i != ra) {
	  SetExplBlock (x, y-i, 0x15);
	} else if (i != ri) {
	  SetExplBlock (x, y-i, 0x14);
	} else if (i == ra) {
	  SetExplBlock (x, y-i, 0x10);
	} else {
	  SetExplBlock (x, y-i, 0x11);
	}
      }
    }
    if (i < ri) {
      *burnOut |= BO_UP;
    }
    if ( (i <= ra) && CheckMazeExtra(x, y-i) ) {
      SetMazeBlock (x, y-i, BTExtraOpen);
    }
  }
  /* down */
  if (0 == (*burnOut & BO_DOWN) ) { 
    for (i = 0; (i <= ra) && CheckMazeOpen (x, y+i); i ++) {
      if (i >= ri) {
	if (i != ri && i != ra) {
	  SetExplBlock (x, y+i, 0x15);
	} else if (i != ri) {
	  SetExplBlock (x, y+i, 0x11);
	} else if (i == ra) {
	  SetExplBlock (x, y+i, 0x10);
	} else {
	  SetExplBlock (x, y+i, 0x14);
	}
      }
    }
    if (i < ri) {
      *burnOut |= BO_DOWN;
    }
    if ( (i <= ra) && CheckMazeExtra(x, y+i) ) {
      SetMazeBlock (x, y+i, BTExtraOpen);
    }
  }
} /* OneExplAt */

/*
 *
 */
static void 
DelExplosion (Explosion *ptr)
{
  Explosion *hilf;
  int i,x,y,r;

  assert (ptr != NULL);

  /* give bomb back to  player */
  if (ptr->player != NULL) {
    ptr->player->bombs++;
    if (ptr->isMorphed) {
      /* remorph player, if he was this bomb */
      ptr->player->morphed = 0;
      ptr->player->x       =  ptr->x      * BLOCK_WIDTH;
      ptr->player->y       = (ptr->y - 1) * BLOCK_HEIGHT;
      ptr->player->num_extras --;
    }
  }
  /* just for convenience */
  x = ptr->x;
  y = ptr->y;
  r = ptr->range;
  /* currect lookup map */
  bombMaze[x][y] = NULL;
  /* one less explosions to worry about */
  numExpl --;
  /* look right for blocks to blast */
  for (i = 0; i <= r; i ++) {
    if (! CheckMazeFree (x + i, y)) {
      BlastExtraBlock (x + i, y);
      break;
    }
  }
  /* look left for blocks to blast */
  for (i = 0; i <= r; i++) {
    if (! CheckMazeFree (x-i, y)) {
      BlastExtraBlock (x-i, y);
      break;
    }
  }
  /* look down for blocks to blast */
  for (i = 0; i <= r; i ++) {
    if (! CheckMazeFree (x, y + i)) {
      BlastExtraBlock (x, y + i);
      break;
    }
  }
  /* look up for blocks to blast */
  for (i = 0; i <= r; i ++) {
    if (! CheckMazeFree (x, y - i)) {
      BlastExtraBlock (x, y - i);
      break;
    }
  }
  /* delete bomb sprite */
  if (ptr->sprite != NULL) {
    DeleteSprite (ptr->sprite);
  }
  /* remove form list */
  if (ptr == explList) {
    explList = ptr->next;
    if (explList == NULL) {
      explEnd = NULL;
    }
  } else {
    for (hilf = explList; hilf->next != NULL; hilf = hilf->next) {
      if (hilf->next == ptr) {
	if (explEnd == ptr) {
	  explEnd = hilf;
	}
	hilf->next = hilf->next->next;
	break;
      }
    }
  }
  /* free memory */
  free (ptr);
} /* DelExplosion */

/*
 * delete all explosions in the game
 */
void
DeleteAllExplosions (void)
{
  Dbg_Out ("delete all explosions\n");
  /* just delete the first element as long as one exists */
  while (NULL != explList) {
    DelExplosion (explList);
  }
  assert (NULL == explList);
  assert (NULL == explEnd);
} /* DeleteAllExplosions */

/*------------------------------------------------------------------------*
 * 
 * Player, wall, and bomb click functions (Garth Denley) 
 *
 *------------------------------------------------------------------------*/

/*
 * no effect (bomb stops)
 */
static void
BombClickNone (Explosion *bomb)
{
  bomb->dir = GoStop;
  bomb->dx  = 0;
  bomb->dy  = 0;
} /* BombClickNone */

/*
 * bomb goes on with inital direction
 */
static void
BombClickInitial (Explosion *bomb)
{
  switch(initialBombDir) {
  case GoStop:
    bomb->dx = 0;
    bomb->dy = 0;
    break;
  case GoRight:
    if (CheckMazeFree (bomb->x+1, bomb->y)) {
      bomb->dx = BOMB_VX;
    }
    bomb->dy = 0;
    break;
  case GoLeft:
    if (CheckMazeFree (bomb->x-1, bomb->y)) {
      bomb->dx = -BOMB_VX;
    }
    bomb->dy = 0;
    break;
  case GoDown:
    bomb->dx = 0;
    if (CheckMazeFree (bomb->x, bomb->y+1)) {
      bomb->dy = BOMB_VY;
    }
    break;
  case GoUp:
    bomb->dx = 0;
    if (CheckMazeFree (bomb->x, bomb->y-1)) {
      bomb->dy = -BOMB_VY;
    }
    break;
  }
  bomb->dir = initialBombDir;
} /* BombClickInitial */

/* 
 * bomb goes thru 
 */
static void
BombClickThru (Explosion *bomb)
{
} /* BombClickThru */

/* 
 * snooker bombs  
 */
static void
BombClickSnooker (Explosion *bomb)
{
  int dir;

  dir = bomb-> dir;

  bomb->dir = GoStop;
  bomb->dx = 0;
  bomb->dy = 0;

  switch (dir) {
  case GoUp:    MoveBomb (bomb->x,   bomb->y-1, dir); break;
  case GoLeft:  MoveBomb (bomb->x-1, bomb->y,   dir); break;
  case GoDown:  MoveBomb (bomb->x,   bomb->y+1, dir); break;
  case GoRight: MoveBomb (bomb->x+1, bomb->y,   dir); break;
  }
} /* BombClickSnooker */

/* 
 * contact bombs 
*/
static void 
BombClickContact (Explosion *bomb)
{
  bomb->dir   = GoStop;
  bomb->dx    = 0;
  bomb->dy    = 0;
  bomb->count = 0;
} /* BombClickContact */

/* 
 * clockwise bombs 
 */
static void 
BombClickClockwise (Explosion *bomb)
{
  static BMDirection turnClockwise[MAX_DIR] = {
    GoStop, GoRight, GoUp, GoLeft, GoDown, GoDefault
  };
  
  bomb->dx  = 0;
  bomb->dy  = 0;
  bomb->dir = turnClockwise[bomb->dir];
} /* BombClickClockwise */

/* 
 * anticlockwise bombs 
 */
static void 
BombClickAnticlockwise (Explosion *bomb)
{
  static BMDirection turnAnticlockwise[MAX_DIR] = {
    GoStop, GoLeft, GoDown, GoRight, GoUp, GoDefault
  };

  bomb->dx  = 0;
  bomb->dy  = 0;
  bomb->dir = turnAnticlockwise[bomb->dir];
} /* BombClickAnticlockwise */

/* 
 * randomdir bombs 
 */
static void 
BombClickRandomdir (Explosion *bomb)
{
  bomb->dx  = 0;
  bomb->dy  = 0;
  bomb->dir = (BMDirection) (GameRandomNumber(4) + 1);
} /* BombClickRandomdir */

/* 
 * rebound bombs 
 */
static void 
BombClickRebound (Explosion *bomb)
{
  static BMDirection turnOpposite[MAX_DIR] = {
    GoStop, GoDown, GoRight, GoUp, GoLeft, GoDefault
  };

  bomb->dir = turnOpposite[bomb->dir];
} /* BombClickRebound */

/* 
 * 
 */
static void 
HauntBomb (Explosion *ptr)
{
  assert (ptr != NULL);

  if (ptr->dir == GoStop) {
    if (! CheckPlayerNear (ptr->x, ptr->y) ) {
      switch (GameRandomNumber(4)) {
      case 0:
	ptr->dir = GoUp;
	ptr->dx  = 0;
	break;
      case 1:
	ptr->dir = GoDown;
	ptr->dx  = 0;
	break;
      case 2:
	ptr->dir = GoLeft;
	ptr->dy  = 0;
	break;
      case 3:
	ptr->dir = GoRight;
	ptr->dy  = 0;
	break;
      }
      SND_Play (SND_HAUNT, SOUND_MIDDLE_POSITION);
    }
  }
} /* HauntKick */

/*
 * move bomb one step upwards 
 */
static void
MoveBombUp (Explosion *ptr)
{
  assert (ptr != NULL);

  if  ( (ptr->dy == 0) && 
	! CheckMazeFree (ptr->x, ptr->y-1) )  {
    (*doWallClick)(ptr);
  } else if ( (ptr->dy <= 0) && CheckBomb(ptr->x,ptr->y-1) ) {
    (*doBombClick)(ptr);
  } else {
    ptr->dy -= BOMB_VY;
    if (ptr->dy <= -BLOCK_HEIGHT/2) {
      SND_Play (SND_SLIDE, (ptr->x * BLOCK_WIDTH) / (PIXW / MAX_SOUND_POSITION));
      bombMaze[ptr->x][ptr->y] = NULL;
      ptr->dy += BLOCK_HEIGHT;
      ptr->y  -= 1;
      bombMaze[ptr->x][ptr->y] = ptr;
    }
  }
} /* MoveBombUp */

/*
 * move bomb one step downwards
 */
static void
MoveBombDown (Explosion *ptr)
{
  assert (ptr != NULL);

  if ( (ptr->dy == 0) && 
       ! CheckMazeFree(ptr->x, ptr->y+1) ) {
    (*doWallClick)(ptr);
  } else if ( (ptr->dy >= 0) && CheckBomb(ptr->x,ptr->y+1)) {
    (*doBombClick)(ptr);
  } else {
    ptr->dy += BOMB_VY;
    if (ptr->dy >= BLOCK_HEIGHT/2) {
      SND_Play (SND_SLIDE, (ptr->x * BLOCK_WIDTH) / (PIXW / MAX_SOUND_POSITION));
      bombMaze[ptr->x][ptr->y] = NULL;
      ptr->dy -= BLOCK_HEIGHT;
      ptr->y  += 1;
      bombMaze[ptr->x][ptr->y] = ptr;
    }
  }
} /* MoveBombDown */

/* 
 * move bomb one step right
 */
static void
MoveBombRight (Explosion *ptr)
{
  assert (ptr != NULL);

  if ( (ptr->dx == 0) && 
       ! CheckMazeFree(ptr->x+1, ptr->y) ) {
    (*doWallClick)(ptr);
  } else if ( (ptr->dx >= 0) && CheckBomb(ptr->x+1,ptr->y)) {
    (*doBombClick)(ptr);
  } else {
    ptr->dx += BOMB_VX;
    if (ptr->dx >= BLOCK_WIDTH/2) {
      SND_Play (SND_SLIDE, (ptr->x * BLOCK_WIDTH) / (PIXW / MAX_SOUND_POSITION));
      bombMaze[ptr->x][ptr->y] = NULL;
      ptr->dx -= BLOCK_WIDTH;
      ptr->x  += 1;
      bombMaze[ptr->x][ptr->y] = ptr;
    }
  }
} /* MoveBombRight */

/*
 * move bomb one step left
 */
void
MoveBombLeft (Explosion *ptr)
{
  assert (ptr != NULL);

  if ( (ptr->dx == 0) && 
       ! CheckMazeFree(ptr->x-1, ptr->y) ) {
    (*doWallClick)(ptr);
  } else if ( (ptr->dx <= 0) && CheckBomb(ptr->x-1,ptr->y)) {
    (*doBombClick)(ptr);
  } else {
    ptr->dx -= BOMB_VX;
    if (ptr->dx <= -BLOCK_WIDTH/2) {
      SND_Play (SND_SLIDE, (ptr->x * BLOCK_WIDTH) / (PIXW / MAX_SOUND_POSITION));
      bombMaze[ptr->x][ptr->y] = NULL;
      ptr->dx += BLOCK_WIDTH;
      ptr->x  -= 1;
      bombMaze[ptr->x][ptr->y] = ptr;
    }
  }
} /* MoveBombLeft */

/*
 * work all bombs in list
 */
void 
DoBombs (void)
{
  Explosion *ptr;
  int bombToHaunt = - 1;

  /* determine if any bomb is haunted */
  if (HAUNT_NONE != hauntFactor) {
    bombToHaunt = GameRandomNumber (hauntFactor);
    if (bombToHaunt > 6) {
      bombToHaunt = -1;
    }
  }

  for (ptr = explList; ptr != NULL; ptr = ptr->next) {
    if (ptr->count == 0) {
      /* Bomb is just exploding */
      if (NULL != ptr->sprite) {
	ptr->anime = MAX_BOMB_ANIME -1;
	SetSpriteAnime (ptr->sprite, ptr->anime);
	SetSpriteMode (ptr->sprite, SPM_MAPPED);
	MoveSprite (ptr->sprite, ptr->x*BLOCK_WIDTH, ptr->y*BLOCK_HEIGHT);
      }
    } else if (ptr->count < 0) {
      /* Bomb has not yet exploded */
      /* haunt bomb */
      if (HAUNT_NONE != hauntFactor) {
	if (0 == bombToHaunt) {
	  HauntBomb (ptr);
	}
	bombToHaunt --;
      }
      /* move bomb */
      switch(ptr->dir) {
      case GoUp:    MoveBombUp (ptr); break;
      case GoDown:  MoveBombDown (ptr); break;
      case GoRight: MoveBombRight (ptr); break;
      case GoLeft:  MoveBombLeft (ptr); break;
      default:      break;
      }
      /* move bomb sprite */
      if (NULL != ptr->sprite) {
	MoveSprite (ptr->sprite, ptr->x*BLOCK_WIDTH + ptr->dx, ptr->y*BLOCK_HEIGHT + ptr->dy);
      }
      /* if player is morphed */
      if (ptr->isMorphed && NULL != ptr->player) {
	ptr->player->x = ptr->dx + ptr->x * BLOCK_WIDTH;
	ptr->player->y = ptr->dy + (ptr->y-1) * BLOCK_HEIGHT;
	MoveSprite (ptr->player->sprite, ptr->player->x, ptr->player->y + BASE_Y);
      }
      /* check if bomb is on an explosion */
      if (CheckExplosion(ptr->x, ptr->y)) {
	ptr->count = 0;
      }
      /* handle bomb animation and blinking */
      if (ptr->sprite != NULL) {
	if (ptr->blink + ptr->count == 0) {
	  SetSpriteMode (ptr->sprite, SPM_MAPPED | SPM_MASKED);
	  ptr->blink = ptr->blink >> 1;
	} else {
	  SetSpriteMode (ptr->sprite, SPM_MAPPED);
	}
	if (ptr->count == ptr->nextAnime) {
	  ptr->anime ++;
	  ptr->nextAnime += curBombTime / (MAX_BOMB_ANIME - 1);
	  SetSpriteAnime (ptr->sprite, ptr->anime);
	}
      }
      /* Bomb malfunction, random or illness */
      if ( (ptr->count == -3) 
	   && (ptr->malfunction || (GameRandomNumber (BOMB_ERROR_PROB) == 0) ) ) {
	ptr->malfunction = 0;
	ptr->count = -BOMB_TIME *(2 + GameRandomNumber (BOMB_DELAY));
	ptr->blink = (BOMB_TIME >>1);
      }
    }
  }
} /* DoBombs */

/* 
 * ignite all bombs of one given player
 */
int 
IgnitePlayersBombs (BMPlayer *ps)
{
  Explosion *ptr;
  int numberOfBombs = 0;

  for (ptr = explList; ptr != NULL; ptr = ptr->next) {
    if (ptr->count < 0) {
      /* draw a bomb */
      if ( ptr->player == ps ) {
	ptr->count = 0;
	numberOfBombs ++;
      }
    }
  }
  return numberOfBombs;
} /* IgnitePlayersBombs */

/* 
 * ignite all bombs 
 */
int 
IgniteAllBombs (BMPlayer *ps)
{
  Explosion *ptr;
  int numberOfBombs = 0;

  for (ptr = explList; ptr != NULL; ptr = ptr->next) {
    if (ptr->count < 0) {
      ptr->count = 0;
      numberOfBombs ++;
    }
  }
  return numberOfBombs;
} /* IgniteAllBombs */

/*
 * ignite bombs on explosions
 */
void 
IgniteBombs (void)
{
  Explosion *ptr;

  for (ptr = explList; ptr != NULL; ptr = ptr->next) {
    if (ptr->count < 0) {
      /* draw a bomb */
      if (CheckExplosion(ptr->x, ptr->y)) {
	ptr->count = 0;
      }
    }
  }
} /* IgniteBombs */

/* 
 * handle all explosions in list
 */
void 
DoExplosions (void)
{
  Explosion *ptr, *nextPtr;
  register int hilf;

  ptr = explList;
  while (ptr != NULL) {
    nextPtr = ptr->next;
    /* check if bomb is exploding */
    if (ptr->count >= 0) {
      /* hide morphed player's eyes */
      if (ptr->count == 0 &&
	  ptr->isMorphed &&
	  ptr->player != NULL) {
	ptr->player->morphed = 3;
      }
      /* check if bomb has burned out */
      if ( (ptr->burnOut == BO_TOTAL) ||
	   (ptr->count  >= (2*ptr->range + 2) ) ) {
	DelExplosion(ptr);
	ptr = nextPtr;
	continue;
      } else {
	/* get exploding time */
	if ( (hilf = ptr->count) == 0) {
	  /* set any free tiel to burned */
	  if (CheckMazeFree (ptr->x, ptr->y) ) {
	    SetMazeBlock (ptr->x, ptr->y, BTBurned);
	  }
	  /* play according sound */
          if (ptr->range == 1 ||
              ptr->type == BMTfirecracker ||
              ptr->type == BMTfirecracker2) {
	    SND_Play (SND_MINIBOMB, (ptr->x * BLOCK_WIDTH) / (PIXW / MAX_SOUND_POSITION));
	  } else {
	    SND_Play (SND_EXPL, (ptr->x * BLOCK_WIDTH) / (PIXW / MAX_SOUND_POSITION));
	  }
	}
	/* now do the explosion */
	OneExplAt (ptr->x, ptr->y, MIN(ptr->range,hilf), MAX(0,( (hilf) - ptr->range)), &(ptr->burnOut) );
      }
    }
    if ((*doSpecialBombFunction[ptr->type])(ptr)) {
      ptr->count ++;
      if ( 2 == ptr->count && 
	   NULL != ptr->sprite ) {
	DeleteSprite (ptr->sprite);
	ptr->sprite = NULL;
      }
    }
    ptr = nextPtr;
  }
} /* DoExplosions */

/*------------------------------------------------------------------------*
 * 
 * Special bomb code (Garth Denley) 
 *
 *------------------------------------------------------------------------*/

/* 
 * Used to spread an explosion out 
 */
static void 
SpreadExplosion (int lx, int ly, int range, int type, int typeExtr)
{
  if ( (lx < MAZE_W) && 
       (lx > -1) && 
       (ly < MAZE_H) && 
       (ly > -1) && 
       ! CheckMazeSolid (lx, ly) ) {
    NewExplosion (NULL, lx, ly, range, XBFalse, XBFalse, type, typeExtr, GoStop);
  }
} /* SpreadExplosion */

/*
 *
 */
static void
MoveBlockFromTo (int sx, int sy, int dx, int dy)
{
  if ( (sx > 0) && (sx < MAZE_W-1) && 
       (sy > 0) && (sy < MAZE_H-1) && 
       (dx > 0) && (dx < MAZE_W-1) && 
       (dy > 0) && (dy < MAZE_H-1) && 
       CheckMazeWall(sx, sy) && 
       CheckMazeFree(dx, dy) ) {
    SetMazeBlock (sx, sy, BTFree);
    SetMazeBlock (dx, dy, BTBlock);
    KillPlayerAt (dx, dy);
    DeleteBombAt (dx, dy);
  }
} /* MoveBlockFromTo */

/*
 * special bomb functions
 */

/*
 *
 */
static XBBool
SpecialBombNormal (Explosion *ptr)
{
  return XBTrue;
}

/* 
 * napalm bomb 
*/
static XBBool
SpecialBombNapalm (Explosion *ptr)
{
  int i;

  if (ptr->count == -1) {
    ptr->type = BMTnormal;
    for (i= -2; i<=2; i++) {
      SpreadExplosion (ptr->x+i, ptr->y, ptr->range/(ABS(i)+1), BMTblastnow, 0);
      SpreadExplosion (ptr->x, ptr->y+i, ptr->range/(ABS(i)+1), BMTblastnow, 0);
    }
  }
  return XBTrue;
} /* SpecialBombNapalm */

/* 
 * firecracker 
 */
static XBBool
SpecialBombFirecracker (Explosion *ptr)
{
  int i;
  int nasty;

  if (ptr->count >= 1) {
    if (ptr->type == BMTfirecracker && 
	GameRandomNumber (10) == 0) {
      ptr->typeExtr = -5;
    }
    nasty = ptr->typeExtr;
    for (i= -1; i<=1; i++) {
      if (nasty < 2 || 0 == GameRandomNumber (1 + nasty) ) {
	SpreadExplosion (ptr->x+i, ptr->y, 1, BMTfirecracker2, nasty + 1);
      }
      if (nasty < 2 || 0 == GameRandomNumber (1 + nasty) ) {
	SpreadExplosion (ptr->x, ptr->y+i, 1, BMTfirecracker2, nasty + 1);
      }
    }
    ptr->type = BMTnormal;
  }

  return XBTrue;
} /* SpecialBombFirecracker */

/* 
 * construction 
 */
static XBBool
SpecialBombConstruction (Explosion *ptr)
{
  int x, y;

  if (ptr->count == 1) {
    x = ptr->x;
    y = ptr->y;
    DelExplosion(ptr);
    if (! CheckPlayerNear (x,y) ) {
      SetMazeBlock (x, y, BTExtra);
      SetBlockExtra (x, y, BTFree);
    }
    return XBFalse;
  }
  return XBTrue;
} /* SpecialBombConstruction */

/*
 * fungus 
 */
static XBBool
SpecialBombFungus (Explosion *ptr)
{
  int i, x, y;

  x = ptr->x;
  y = ptr->y;
  if (ptr->count == (-curBombTime)*3/5) {
    for (i = -1; i<=1; i++) {
        SpreadExplosion(x+i,y,1,BMTfungus,0);
        SpreadExplosion(x,y+i,1,BMTfungus,0);
      }
  }
  return XBTrue;
} /* SpecialBombFungus */

/* 
 * threebombs 
 */
static XBBool
SpecialBombThreebombs (Explosion *ptr)
{
  if (ptr->count == -curBombTime) {
    SpreadExplosion(ptr->x-2,ptr->y,ptr->range,defaultBMT,0);
    SpreadExplosion(ptr->x+2,ptr->y,ptr->range,defaultBMT,0);
  }
  return XBTrue;
} /* SpecialBombThreebombs */

/* 
 * grenade 
 */
static XBBool
SpecialBombGrenade (Explosion *ptr)
{
  int i, j;

  if (ptr->range > 0) {
    if (ptr->count == -1) {
      if (ptr->range == 1) {
	SpreadExplosion(ptr->x-1, ptr->y-1, 0, BMTblastnow, 0);
	SpreadExplosion(ptr->x+1, ptr->y-1, 0, BMTblastnow, 0);
	SpreadExplosion(ptr->x-1, ptr->y+1, 0, BMTblastnow, 0);
	SpreadExplosion(ptr->x+1, ptr->y+1, 0, BMTblastnow, 0);
      } else {
	for (i = -((ptr->range)-1); i<=((ptr->range)-1); i++) {
	  for (j = -((ptr->range)-1); j<=((ptr->range)-1); j++) {
	    SpreadExplosion (ptr->x+i, ptr->y+j, 1, BMTblastnow, 0);
	  }
	}
      }
    }
  }
  return XBTrue;
} /* SpecialBombGrenade */

/* 
 * trianglebombs 
*/
static XBBool
SpecialBombTrianglebombs (Explosion *ptr)
{
  if (ptr->count == -curBombTime + 2) {
    int i,j;
    
    i = GameRandomNumber (2) * 4 - 2;
    j = GameRandomNumber (2) * 4 - 2;
    SpreadExplosion (ptr->x+i, ptr->y,   ptr->range, BMTnormal, 0);
    SpreadExplosion (ptr->x,   ptr->y+j, ptr->range, BMTnormal, 0);
    ptr->type = BMTnormal;
  }
  return XBTrue;
} /* SpecialBombTrianglebombs */

/* desctruction */
static XBBool
SpecialBombDestruction (Explosion *ptr)
{
  int i, x, y;

  x = ptr->x;
  y = ptr->y;
  if (ptr->count == 1) {
    DelExplosion(ptr);
    for (i = -1; i<=1; i++) {
      if ( (x+i < (MAZE_W-1)) && 
	   CheckMaze(x+i,y) && 
	   (x+i > 0) ) {
	SetMazeBlock(x+i,y,BTFree);
      }
      if ( (y+i < (MAZE_H-1)) && 
	   CheckMaze(x,y+i) && 
	   (y+i > 0) ) {
	SetMazeBlock(x,y+i,BTFree);
      }
    }
    return XBFalse;
  }
  return XBTrue;
} /* SpecialBombDestruction */

/* 
 * renovation 
*/
static XBBool
SpecialBombRenovation (Explosion *ptr)
{
  int x, y;

  x = ptr->x;
  y = ptr->y;
  if (ptr->count == 1) {
    MoveBlockFromTo (x-1, y,   x-2, y  );
    MoveBlockFromTo (x+1, y,   x+2, y  );
    MoveBlockFromTo (x,   y-1, x,   y+1);
    MoveBlockFromTo (x-1, y-1, x,   y+2);
  }
  return XBTrue;
} /* SpecialBombdRenovation */

/* 
 * pyro 
*/
static XBBool
SpecialBombPyro (Explosion *ptr)
{
  int x, y, k;

  if (ptr->count == 1) {
    for (k=0;k<5;k++) {
      x = ptr->x + GameRandomNumber (3) - 1;
      y = ptr->y + GameRandomNumber (3) - 1;  
      if ( ! bombMaze[x][y] && 
	   CheckMazeFree (x,y) ) {
	SpreadExplosion (x, y, 1, BMTpyro2, 0);
	break;
      }
    }
  }
  return XBTrue;
} /* SpecialBombPyro */

/*
 *
 */
static XBBool 
NewExplosion (BMPlayer *player, int x, int y, int range, XBBool remote_controlled, XBBool malfunction, BMBombType type, 
	      int typeExtr, BMDirection initialdir)
{
  Explosion *newExpl;

  /* check if there is already a bomb in this tile */
  if (NULL != bombMaze[x][y]) {
    return XBFalse;
  }
  /* mark for redraw */
  MarkMazeTile (x,y);
  /* alloc data */
  newExpl = (Explosion *) calloc(1,sizeof(Explosion));
  /* put in lookup map */
  bombMaze[x][y] = newExpl;
  /* put in explosion list */
  numExpl ++;
  if (explList == NULL) {
    explList = newExpl;
  } else {
    explEnd->next = newExpl;
  }
  explEnd   = newExpl;
  /* set values */
  newExpl->next        = NULL;
  newExpl->player      = player;
  newExpl->x           = x;
  newExpl->y           = y;
  newExpl->dx          = 0;
  newExpl->dy          = 0;
  newExpl->malfunction = malfunction;
  newExpl->isMorphed   = (NULL != player) ? player->morphed : XBFalse;
  newExpl->dir         = (initialdir == GoDefault) ? initialBombDir : initialdir;
  /* set type */
  switch (type) {
  case BMTdefault: type = defaultBMT; break;
  case BMTspecial: type = specialBMT; break;
  case BMTevil:    type = evilBMT;    break;
  default:                            break;
  }
  newExpl->type = type;
  /* Random bomb ! */
  if (newExpl->type == BMTrandom) {
    switch (GameRandomNumber (5)) {
    case 0: newExpl->type = BMTnapalm;      break;
    case 1: newExpl->type = BMTfirecracker; break;
    case 2: newExpl->type = BMTgrenade;     break;
    case 3: newExpl->type = BMTfungus;      break;
    case 4: newExpl->type = BMTpyro;        break;
    }
  }
  /* extry type info */
  newExpl->typeExtr = typeExtr;
  /* these are "nasty bombs"  */
  if (type != BMTclose) {
    switch(newExpl->dir) {
    case GoDown:
      if(CheckMazeFree(newExpl->x,newExpl->y+1)) {
	newExpl->dy = BOMB_VY;
      }
      break;
    case GoUp:
      if(CheckMazeFree(newExpl->x,newExpl->y-1)) {
	newExpl->dy = -BOMB_VY;
      }
      break;
    case GoLeft:
      if(CheckMazeFree(newExpl->x-1,newExpl->y)) {
	newExpl->dx = -BOMB_VX;
      }
      break;
      
    case GoRight:
      if(CheckMazeFree(newExpl->x+1,newExpl->y)) {
	newExpl->dx = BOMB_VX;
      }
      break;

    default:
      break;
    }
  }
  /* set range according to type */
  switch (type) {
  case BMTfirecracker:
  case BMTfungus:
  case BMTpyro:     
    newExpl->range = 1; 
    break;
  case BMTconstruction:
    newExpl->range = 0;
    break;
  case BMTgrenade:
    newExpl->range = range / 2;
    break;
  default:
    newExpl->range = range;
  }
  /* bomb counter */
  if ( (type == BMTblastnow) || 
       (type == BMTfirecracker2) || 
       (type == BMTpyro2) ) {
    /* bombs which explode immediately */
    newExpl->count = 0;
  } else if (remote_controlled) {
    /* remote controled bombs */
    newExpl->count     = -GAME_TIME;
    newExpl->anime     = MAX_BOMB_ANIME-2;
    newExpl->nextAnime = 1;
  } else if (type == BMTshort) {
    /* fast bombs */
    newExpl->count     = -curBombTime/4;
    newExpl->anime     = 3*MAX_BOMB_ANIME/4;
    newExpl->nextAnime = 1 + curBombTime/(MAX_BOMB_ANIME - 1) - curBombTime/4;
  } else {
    /* standard bombs */
    newExpl->count = -curBombTime;
    newExpl->anime = 0;
    newExpl->nextAnime = 1 + curBombTime/(MAX_BOMB_ANIME - 1) - curBombTime;
  }
  /* init blilinking */
  newExpl->blink = BOMB_TIME >> 1;
  /* create bomb sprite */
  if ( (type == BMTblastnow) || 
       (type == BMTfirecracker2) || 
       (type == BMTpyro2) ) {
    /* instant explosion => no sprite */
    newExpl->sprite = NULL;
  } else if ( (range ==1) || 
	      (type == BMTfirecracker) ||
	      (type == BMTfungus) ||
	      (type == BMTpyro) ) {
    /* these bombs are shown as mini bombs */
    newExpl->sprite = CreateBombSprite (BB_MINI, newExpl->x*BLOCK_WIDTH, newExpl->y*BLOCK_HEIGHT,
					newExpl->anime, SPM_MAPPED);
  } else {
    /* just the normalbomb sprite */
    newExpl->sprite = CreateBombSprite (BB_NORMAL, newExpl->x*BLOCK_WIDTH, newExpl->y*BLOCK_HEIGHT,
					newExpl->anime, SPM_MAPPED);
  }
  return XBTrue;
} /* NewExplosion */

/*
 *
 */
XBBool
NewPlayerBomb (BMPlayer *ps, BMBombType type)
{
  return NewExplosion (ps, (ps->x + BLOCK_WIDTH/2)/BLOCK_WIDTH,
		       (ps->y + BLOCK_HEIGHT + BLOCK_HEIGHT/2)/BLOCK_HEIGHT,
		       (ps->illness == IllMini) ? 1 : ps->range,
		       (ps->remote_control > 0),
		       (ps->illness == IllMalfunction),
		       type, 0, GoDefault );
} /* NewPlayerBomb */

/*
 *
 */
XBBool
NewEvilBomb (int x, int y)
{
  return NewExplosion (NULL, x, y, 3, XBFalse, XBFalse, evilBMT, 0, GoDefault);
} /* NewEvilBomb */

/* 
 * check if any moving stun (hit) any player
 */
void 
StunPlayers (BMPlayer *ps, int numPlayer)
{
  Explosion *ptr;
  int player;
  unsigned clickFlags;

  for (ptr = explList; ptr != NULL; ptr = ptr->next) {
    /* only if bomb is moving */
    if (ptr->dir != GoStop) {
      clickFlags = 0;
      for (player = 0; player < numPlayer; player ++) {
	/* check if any is vulnerable player is in range */
	if ( (ps[player].invincible == 0) && 
	     (ps[player].morphed == 0) &&
	     (ABS(ptr->x*BLOCK_WIDTH + ptr->dx - ps[player].x ) < BOMB_STUN_X ) && 
	     (ABS(ptr->y*BLOCK_HEIGHT + ptr->dy - ps[player].y - BLOCK_HEIGHT ) < BOMB_STUN_Y ) ) {
	  /* we need to correct some graphics here */
	  if (ptr->dx == 0) {
	    if (ptr->dy < 0) {
	      MarkMazeTile(ptr->x, ptr->y-1);
	    }
	    if (ptr->dy > 0) {
	      MarkMazeTile(ptr->x, ptr->y+1);
	    }
	  }
	  if (ptr->dy == 0) {
	    if (ptr->dx < 0) {
	      MarkMazeTile(ptr->x-1, ptr->y-1);
	    }
	    if (ptr->dx > 0) {
	      MarkMazeTile(ptr->x+1, ptr->y-1);
	    }
	  }
	  /* mark that player has been hit */
	  clickFlags |= (1<<player);
	}
      }
      /* do player click after all players are checked */
      if (clickFlags) {
	for (player = 0; player < numPlayer; player ++) {
	  /* has player been hit ... */
	  if (clickFlags & (1<<player)) {
	    /* ... stun player */
	    if (0 == ps[player].stunned) {
	      ps[player].stunned = STUN_TIME;
	      SND_Play (SND_STUN, ps[player].x / (PIXW/MAX_SOUND_POSITION) );
	    }
	  }
	}
	/* do bomb-player click */
	(*doPlayerClick)(ptr);
      }
    }
  }
} /* StunPlayers */

/* 
 * check if there is a bomb on a tile
 */
XBBool
CheckBomb (int x, int y)
{
  return (bombMaze[x][y] != NULL) && (bombMaze[x][y]->count < 0);
} /* CheckBomb */

/* 
 * how many bombs and explosions do we have
 */
int 
NumberOfExplosions (void)
{
  return numExpl;
} /* NumberOfExplosions */

/*
 * delete a bomb at a given position
 */
void 
DeleteBombAt (int x, int y)
{
  if (NULL != bombMaze[x][y]) {
    DelExplosion (bombMaze[x][y]);
  }
} /* DeleteBombAt */

/* 
 * move a bomb 
 */
void 
MoveBomb (int x, int y, int dir)
{
  Explosion *ptr;
  
  if(NULL != (ptr = bombMaze[x][y]) ) {
    if (ptr->dir == GoStop) {
      ptr->dir = dir;
      switch(dir) {
      case GoUp:
      case GoDown:
	ptr->dx = 0;
	break;
      case GoLeft:
      case GoRight:
	ptr->dy = 0;
	break;
      }
    }
  }
} /* MoveBomb */

/*
 * check where we can savely distribute extras
 */
int 
CheckDistribExpl (unsigned *distExtra, int freeBlocks) 
{
  Explosion *ptr;
  int x, y, ra;

  /* Go through explosions */
  for (ptr = explList; ptr != NULL; ptr = ptr->next) {
    if (ptr->count >=0) {
      ra = MIN(ptr->range, ptr->count + 1);
      /* right */
      for (x = ptr->x; x <= (ptr->x + ra) && x < MAZE_W; x ++) {
	if (distExtra[ptr->y] & (1 << x)) {
	  freeBlocks --;
	}
	distExtra[ptr->y] &= ~ (1 << x);
      }
      /* left */
      for (x = ptr->x; x >= (ptr->x - ra) && x >= 0; x --)	{
	if (distExtra[ptr->y] & (1 << x)) {
	  freeBlocks --;
	}
	distExtra[ptr->y] &= ~ (1 << x);
      }
      /* down */
      for (y = ptr->y; y <= (ptr->y + ra) && y < MAZE_H; y ++) {
	if (distExtra[y] & (1 << ptr->x)) {
	  freeBlocks --;
	}
	distExtra[y] &= ~ (1 << ptr->x);
      }
      for (y = ptr->y; y >= (ptr->y - ra) && y >= 0; y --)	{
	if (distExtra[y] & (1 << ptr->x) ) {
	  freeBlocks --;
	}
	distExtra[y] &= ~ (1 << ptr->x);
      }
    }
  }
  return freeBlocks;
} /* CheckDistribExpl */

/* 
 * public function do_air (Garth Denley)
 * Shoots bombs away if within 2 square radius 
 * Direction based on angle from bomb 
 */
void 
DoAir (BMPlayer *ps)
{
  Explosion *ptr;
  int x, y, ex, ey;

  assert (NULL != ps);
  for (ptr = explList; ptr != NULL; ptr = ptr->next) {
    if (ptr->dir == GoStop)  {
      x  = (ptr->x)   * BLOCK_WIDTH;
      y  = (ptr->y-1) * BLOCK_HEIGHT;
      ex = x - ps->x;
      ey = y - ps->y;
      if ( ABS (ex) < 2*BLOCK_WIDTH && 
	   ABS (ey) < 2*BLOCK_HEIGHT && 
	   (ex != 0 || ey != 0) ) {
        if (ABS (ex) * BLOCK_HEIGHT >= ABS(ey) * BLOCK_WIDTH)  {
          if (ex < 0) {
	    ptr->dir = GoLeft; 
	  } else { 
	    ptr->dir = GoRight;
	  }
        } else {
          if (ey < 0) {
	    ptr->dir = GoUp; 
	  } else {
	    ptr->dir = GoDown;
	  }
        }
      }
    }
  }
} /* DoAir */

/*
 *
 */
void
DoNastyWalls (int gameTime)
{
  int x, y;
  int dir = GoStop;

  if (gameTime >= nextNasty) {
    if (gameTime >= nextNasty + NASTY_INC) {
      nextNasty += NASTY_INC*2;
    }
    if (GameRandomNumber (gentleNasty) < gameTime) {
      dir = GoUp + GameRandomNumber (4);
      switch (dir) {
      case GoUp:
	x   = GameRandomNumber (MAZE_W-2) + 1;
	y   = MAZE_H-2;
	break;
      case GoLeft:
	x   = MAZE_W-2;
	y   = GameRandomNumber (MAZE_H-2) + 1;
	break;
      case GoDown:
	x   = GameRandomNumber (MAZE_W-2) + 1;
	y   = 1;
	break;
      case GoRight:
	x   = 1;
	y   = GameRandomNumber (MAZE_H-2) + 1;
	break;
      default:
	return;
      }
      NewExplosion (NULL, x, y, rangeNasty, XBFalse, XBFalse, BMTclose, 0, dir);
    }
  }
} /* DoNastyWalls */

/*
 * end of file bomb.c
 */
