/*
 * file game_server.c - run game as server
 *
 * $Id: game_server.c,v 1.9 2001/06/27 06:04:25 xblast Exp $
 *
 * Program XBLAST 
 * (C) by Oliver Vogel (e-mail: m.vogel@ndh.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will be entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "game_server.h"

#include "atom.h"
#include "cfg_level.h"
#include "bomb.h"
#include "demo.h"
#include "game.h"
#include "geom.h"
#include "intro.h"
#include "level.h"
#include "server.h"
#include "status.h"

/*
 * local variables
 */
static CFGGame      serverGame;
static PlayerAction serverAction[MAX_PLAYER];
static XBBool       playerLinked[MAX_PLAYER];

/*
 *
 */
static void 
InitPlayerLink (void)
{
  int i;

  assert (serverGame.players.num <= MAX_PLAYER);
  for (i = 0; i < serverGame.players.num; i ++) {
    playerLinked[i] = (serverGame.players.host[i] != XBPH_Server && 
		       serverGame.players.host[i] != XBPH_Local  && 
		       serverGame.players.host[i] != XBPH_Demo);
  }
  for (; i < MAX_PLAYER; i ++) {
    playerLinked[i] = XBFalse;
  }
} /* InitPlayerLink */

/*
 *
 */
static XBBool
UpdatePlayerLink (void)
{
  int    i;
  XBBool result = XBFalse;

  for (i = 0; i < serverGame.players.num; i ++) {
    if (serverGame.players.host[i] != XBPH_Server && 
	serverGame.players.host[i] != XBPH_Local  &&
	! playerLinked[i] &&
	! player_stat[i].in_active) {
      player_stat[i].in_active = XBTrue;
      result = XBTrue;
    }
  }
  return result;
} /* UpdatePlayerLink */

/*
 * wait for message from all clients
 */
static void 
WaitForClientEvent (XBNetworkEvent waitEvent, XBBool needFlush)
{
  int            i, num;
  unsigned       id;
  XBEventData    eData;
  XBNetworkEvent netEvent;
  XBBool         playerWait[MAX_PLAYER];
  
  /* for which players do we wait */
  memcpy (playerWait, playerLinked, sizeof (playerWait));
  /* set event handling */
  GUI_SetTimer (FRAME_TIME, XBTrue);
  GUI_SetKeyboardMode (KB_NONE);
  GUI_SetMouseMode (XBFalse);
  
  do {
    /* how many player are not ready */
    num = 0;
    for (i = 0; i < serverGame.players.num; i ++) {
      if (playerWait[i]) {
	num ++;
      } 
    }
    /* update window */
    GameUpdateWindow ();
    /* wait for next event */
    if (XBE_TIMER == GUI_WaitEvent (&eData) ) {
      /* try to flush udp connections */
      if (needFlush) {
	needFlush = Server_FlushPlayerAction ();
      }
      /* check for nework events */
      netEvent = Network_GetEvent (&id);
      if (netEvent == waitEvent ||
	  netEvent == XBNW_Error ||
	  netEvent == XBNW_Disconnected) {
	XBPlayerHost host = XBPH_Client1 + id - 1;
	for (i = 0; i < serverGame.players.num; i ++) {
	  if (serverGame.players.host[i] == host) {
	    playerWait[i] = XBFalse;
	    if (netEvent == XBNW_Error ||
		netEvent == XBNW_Disconnected) {
	      playerLinked[i] = XBFalse;
	      Dbg_Out ("unlink player %d\n", i + 1);
	    }
	  }
	}
      }
    }
  } while (num > 0);
} /* WaitForClientEvent */

/*
 * server wants to synchronize with clients
 */
static void
SyncWithClients (XBNetworkEvent syncEvent, XBBool needFlush, XBBool showMsg)
{
  if (showMsg) {
    SetMessage ("Waiting for others ...", XBTrue);
  }
  WaitForClientEvent (syncEvent, needFlush);
  Server_SendSync (syncEvent);
} /* SyncWithClients */

/*
 * insert keys from clients
 */
static void
InsertClientAction (const CFGGamePlayers *cfgPlayers, PlayerAction *serverAction)
{
  int i;

  assert (NULL != cfgPlayers);
  assert (NULL != serverAction);

  for (i = 0; i < cfgPlayers->num; i ++) {
    switch (cfgPlayers->host[i]) {
    case XBPH_Client1: Server_GetPlayerAction (1, i, serverAction + i); break;
    case XBPH_Client2: Server_GetPlayerAction (2, i, serverAction + i); break;
    case XBPH_Client3: Server_GetPlayerAction (3, i, serverAction + i); break;
    case XBPH_Client4: Server_GetPlayerAction (4, i, serverAction + i); break;
    case XBPH_Client5: Server_GetPlayerAction (5, i, serverAction + i); break;
    default:           break;
    }
  }
  Server_ClearPlayerAction ();
} /* InsertClientAction */

/*
 *
 */
static int 
ServerRunLevel (int numActive, const DBRoot *level)
{
  int 	       gameTime;
  int 	       pauseStatus;
  int 	       lastTeam;
  int 	       frameTime;
  const char  *msg;
  XBEventData  eData;

  /* sanity check */
  assert (level != NULL);
  /* necesary inits */
  gameTime    = 0;
  pauseStatus = -1;
  lastTeam    = -1;
  frameTime   = serverGame.setup.frameRate ? 1000/serverGame.setup.frameRate : 0;
  /* start demo recording */
  if (serverGame.setup.recordDemo) {
    DemoInitLevel (DB_Atom (level));
  }
  /* Config level */
  ConfigLevel (level);
  /* show level info */
  if (! LevelIntro (serverGame.players.num, level, XBTrue)) {
    goto Exit;
  }
  /* syncshronize clients and server */
  SyncWithClients (XBNW_SyncLevelIntro, XBFalse, XBTrue);
  /* clean up */
  Server_ClearPlayerAction ();
  Server_ResetPlayerAction ();
  /* init level display */
  LevelBegin (GetLevelName (level));
  /* inti events */
  GUI_SetTimer (frameTime, XBTrue);
  GUI_SetKeyboardMode (KB_XBLAST);
  GUI_SetMouseMode (XBFalse);
  /* now start */
  do {
    /*
     * INGAME mode
     */
    /* ready input */
    ClearPlayerAction (serverAction);
    /* handle all event until timer triggers */
    if (! GameEventLoop (XBE_TIMER, &eData) ) {
      goto Exit;
    }
    /* increment game clock */
    gameTime ++;
    /* handle game turn */
    GameTurn (gameTime, serverGame.players.num, &numActive);
    /* insert any data received from clients */
    InsertClientAction (&serverGame.players, serverAction);
    /* send all data on player actions to clients */
    Server_SendPlayerAction (gameTime, serverAction);
    /* record demo data if neede */
    if (serverGame.setup.recordDemo) {
      DemoRecordFrame (gameTime, serverAction);
    }
    /* evaluate player action */
    (void) GameEvalAction (serverGame.players.num, serverAction);
    /* update window */
    GameUpdateWindow ();
  } while (gameTime < GAME_TIME && 
	   numActive > 0 && 
	   (numActive > 1 || NumberOfExplosions () != 0) );
  /* finisg demo file if needed */
  if (serverGame.setup.recordDemo) {
    DemoFinishLevel (gameTime);
  }
  /* tell client game ist over */
  Server_FinishPlayerAction (gameTime + 1);
  /* wait for clients to acknowledge */
  SyncWithClients (XBNW_SyncLevelResult, XBTrue, XBFalse);
  /* calc result etc */
  msg = LevelResult (gameTime, &lastTeam, serverGame.players.num, level, XBTrue);
  /* show message and winner Animation */
  if (! LevelEnd (serverGame.players.num, lastTeam, msg, XBTrue) ) {
    lastTeam = -1;
  }
  /* clean up */
 Exit:
  FinishLevel ();
  DeleteAllExplosions ();
  /* fade out image */
  DoFade (XBFM_BLACK_OUT, PIXH+1);
  /* thats all */
  return lastTeam;
} /* ServerRunLevel */

/*
 * send level data to clients 
 */
static void
SendLevelToClients (const DBRoot *level)
{
  /* send level data to clients */
  Server_SendLevel (level);
  /* TODO wait for acknowledge */
} /* SendLevelToClients */

/*
 *
 */
void
RunServerGame (void)
{
  const DBRoot  *level;
  int            lastTeam;
  int            i, maxNumWins;
  int            numActive;
  XBBool         initDone  = XBFalse;

  /* get setup */
  if (! RetrieveGame (CT_Remote, atomArrayHost0[0], &serverGame) ) {
    goto Exit;
  }
  /* select levels to plays */
  if (! InitLevels (&serverGame) ) {
    goto Exit;
  }
  /* common inits */
  if (! InitGame (XBPH_Server, CT_Remote, &serverGame, serverAction)) {
    goto Exit;
  }
  initDone = XBTrue;
  /* init connection status */
  InitPlayerLink ();
  /* network sync */
  SyncWithClients (XBNW_SyncEndOfInit, XBFalse, XBFalse);
  /* game loop */
  maxNumWins = 0;
  numActive  = serverGame.players.num;
  do {
    /* load and run next level */
    level    = LoadLevelFile (GetNextLevel ());
    SendLevelToClients (level);
    lastTeam = ServerRunLevel (numActive, level);
    /* check for quick exit */
    if (-1 == lastTeam) {
      goto Exit;
    }
    /* calculate new maxmium # of victories */
    for (i = 0; i < serverGame.players.num; i ++) {
      if (player_stat[i].victories > maxNumWins) {
	maxNumWins = player_stat[i].victories;
      }
    }
    /* sync before showing score board */
    SyncWithClients (XBNW_SyncLevelEnd, XBFalse, XBTrue);
    /* correct number of players */
    UpdatePlayerLink ();
    /* show scores */
    if (! ShowScoreBoard (lastTeam, maxNumWins, serverGame.players.num, player_stat, XBTrue)) {
      goto Exit;
    }
    /* sync after showing score board */
    SyncWithClients (XBNW_SyncScoreboard, XBFalse, XBTrue);
    /* determine number of active players */
    numActive = 0;
    for (i = 0; i < serverGame.players.num; i ++) {
      if (! player_stat[i].in_active) {
	numActive ++;
      }
    }
    Dbg_Out ("%d active players\n", numActive);
  } while (numActive > 1 &&
	   maxNumWins < serverGame.setup.numWins);
  /* close connection to clients */
  Server_SendDisconnectAll ();
  /* and the winner is ... */
  if (maxNumWins >= serverGame.setup.numWins) {
    InitWinner (serverGame.players.num);
    ShowWinner (lastTeam, serverGame.players.num, player_stat);
  } else {
    GUI_ErrorMessage ("Not enough players left in the game");
  }
  /* that's all */
  FinishGame (&serverGame);
  return;

  /*
   * fast exit via Escape key ...
   */
 Exit:
  Dbg_Out ("aborting server game\n");
  Server_SendDisconnectAll ();
  if (initDone) {
    FinishGame (&serverGame);
  }
  return;
} /* StartServerGame */

/*
 * end of file game_server.c
 */
