/*
 * file menu.c - user interface for game setup
 *
 * $Id: menu.c,v 1.54 2001/06/27 20:42:26 xblast Exp $
 *
 * Program XBLAST 
 * (C) by Oliver Vogel (e-mail: m.vogel@ndh.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will be entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "menu.h"

#include "atom.h"
#include "geom.h"
#include "intro.h"
#include "menu_control.h"
#include "menu_game.h"
#include "menu_level.h"
#include "menu_player.h"
#include "menu_extras.h"
#include "menu_layout.h"
#include "snd.h"

/*
 * local constants
 */
#define ANIME_WAVE    9
#define ANIME_HOLD   12
#define ANIME_LENGTH (ANIME_WAVE+ANIME_HOLD)

/*
 * local variables
 */
static XBAtom gameType;
static const CFGPlayerGraphics *audience_ptr[MAX_PLAYER];
static BMSpriteAnimation player_anime[ANIME_WAVE+2*ANIME_HOLD] = {
    SpriteStopDown, SpriteStopDown, SpriteStopDown, 
    SpriteStopDown, SpriteStopDown, SpriteStopDown, 
    SpriteStopDown, SpriteStopDown, SpriteStopDown, 
    SpriteStopDown, SpriteStopDown, SpriteStopDown, 
    SpriteStopDown, SpriteWinner3,  SpriteWinner2, 
    SpriteWinner,   SpriteWinner,   SpriteWinner, 
    SpriteWinner2,  SpriteWinner3,  SpriteStopDown, 
    SpriteStopDown, SpriteStopDown, SpriteStopDown, 
    SpriteStopDown, SpriteStopDown, SpriteStopDown, 
    SpriteStopDown, SpriteStopDown, SpriteStopDown, 
    SpriteStopDown, SpriteStopDown, SpriteStopDown, 
};
/* combo entries for sound mode */
static XBComboEntryList soundModeList[] = {
  { "none",    XBSM_None,    NULL, ATOM_INVALID, },
  { "beep",    XBSM_Beep,    NULL, ATOM_INVALID, },
  { "waveout", XBSM_Waveout, NULL, ATOM_INVALID, },
  { NULL,      0,            NULL, ATOM_INVALID, },
};

/*
 * start local game
 */
static XBBool
ButtonStartLocal (void *par)
{
  gameType = atomLocal;
  return CreateLocalGameMenu (&gameType);
} /* ButtonStartLocal */

/*
 * create a network game
 */
static XBBool
ButtonCreateNet (void *par)
{
  gameType = atomServer;
  return CreateStartNetGameMenu (&gameType);
} /* ButtonStartLocal */

/*
 * join a network game
 */
static XBBool
ButtonJoinNet (void *par)
{
  gameType = atomClient;
  return CreateJoinNetGameMenu (&gameType);
} /* ButtonJoinNet */

/*
 * quit xblast menu
 */
static XBBool
ButtonMainQuit (void *par)
{
  MenuClear ();
  return XBFalse;
} /* ButtonMenuQuit */

/*
 * create the main menu
 */
XBBool
CreateMainMenu (void *par)
{
  int     i;

  gameType = ATOM_INVALID;
  /* setup menu */
  MenuClear ();
  /* Title */
  MenuAddLabel  (TITLE_LEFT, TITLE_TOP, TITLE_WIDTH, "XBlast Main Menu");
  /* Buttons */
  (void) MenuAddHButton (MENU_LEFT, MENU_TOP,            MENU_WIDTH, "Start Local Game",    ButtonStartLocal,    NULL);
  (void) MenuAddHButton (MENU_LEFT, MENU_TOP + 1*CELL_H, MENU_WIDTH, "Create Network Game", ButtonCreateNet,     NULL);
  (void) MenuAddHButton (MENU_LEFT, MENU_TOP + 2*CELL_H, MENU_WIDTH, "Join Network Game",   ButtonJoinNet,       NULL);
  (void) MenuAddHButton (MENU_LEFT, MENU_TOP + 3*CELL_H, MENU_WIDTH, "Options",             CreateOptionsMenu,   NULL);
  (void) MenuAddHButton (MENU_LEFT, MENU_TOP + 4*CELL_H, MENU_WIDTH, "Extras",              CreateExtrasMenu,    NULL);
  /* exit game */
  MenuSetAbort (MenuAddHButton (MENU_LEFT, MENU_BOTTOM,  MENU_WIDTH, "Quit",                ButtonMainQuit,      NULL) );
  /* audience */
  for (i = 0; i < MAX_PLAYER; i ++) {
    audience_ptr[i] = DefaultPlayerGraphics (i);
    MenuAddPlayer (PLAYER_LEFT (i, MAX_PLAYER), PLAYER_TOP, PLAYER_WIDTH, i, audience_ptr +i, 
		   ANIME_LENGTH, player_anime + ANIME_HOLD - 2*i -2);
  }
  /* return and escape */
  MenuSetLinks ();
  /* that's all*/
  return XBFalse;
} /* CreateMainMenu */

/*
 * save changed sound setup
 */
static XBBool
ButtonSaveSound (void *par)
{
  assert (NULL != par);
  /* store setup */
  StoreSoundSetup (par);
  /* reactivate sound setup */
  SND_Finish ();
  RetrieveSoundSetup (par);
  SND_Init (par);
  /* back to options */
  return CreateOptionsMenu (NULL);
} /* ButtonSaveSoundSetup */

/*
 * menu for sound setup 
 */ 
static XBBool
CreateSoundMenu (void* par)
{
  static CFGSoundSetup soundSetup;

  RetrieveSoundSetup (&soundSetup);
  /* --- */
  MenuClear ();
  /* Title */
  MenuAddLabel  (TITLE_LEFT, TITLE_TOP, TITLE_WIDTH, "Sound Setup");
  /* Buttons */				   
  MenuAddComboInt  (5*CELL_W, MENU_ROW(0), 5*CELL_W, "Mode:",   3*CELL_W, (int *)&soundSetup.mode, soundModeList);
  MenuAddComboBool (5*CELL_W, MENU_ROW(1), 5*CELL_W, "Stereo:", 3*CELL_W, &soundSetup.stereo);
  /* leave menu */
  /* ok and cancel */
  MenuSetAbort   (MenuAddHButton ( 5 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Cancel", CreateOptionsMenu, NULL) );
  MenuSetDefault (MenuAddHButton (17 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Ok",     ButtonSaveSound,   &soundSetup) );
  /* escape */
  MenuSetLinks ();
  /* that's all*/
  return XBFalse;
} /* CreateSoundMenu */

/*
 * create the options menu
 */
XBBool
CreateOptionsMenu (void *par)
{
  MenuClear ();
  /* Title */
  MenuAddLabel  (TITLE_LEFT, TITLE_TOP, TITLE_WIDTH, "XBlast Options Menu");
  /* Buttons */				   
  MenuAddHButton (MENU_LEFT, MENU_ROW(0), MENU_WIDTH, "Players",  CreatePlayerOptionsMenu, NULL);
  MenuAddHButton (MENU_LEFT, MENU_ROW(1), MENU_WIDTH, "Controls", CreateConfigControlMenu, NULL);
  MenuAddHButton (MENU_LEFT, MENU_ROW(2), MENU_WIDTH, "Sound",    CreateSoundMenu,         NULL);
  /* leave menu */
  MenuSetAbort (MenuAddHButton (MENU_LEFT, MENU_BOTTOM, MENU_WIDTH, "Main Menu", CreateMainMenu, NULL) );
  /* escape */
  MenuSetLinks ();
  /* that's all*/
  return XBFalse;
} /* CreateOptionsMenu */

/*
 *
 */
XBPlayerHost
DoMenu (void)
{
  int result;

  GUI_ClearPixmap ();
  /* create main menu */
  (void) CreateMainMenu (NULL);
  /* run menu */
  result = MenuEventLoop ();
  /* fade out */
  if (result) {
    DoFade (XBFM_WHITE_OUT, PIXH + SCOREH);
  } else {
    DoFade (XBFM_BLACK_OUT, PIXH + SCOREH);
  }
  /* thats all */
  return result ?  GetHostType () : XBPH_None;
} /* DoMenu */

/*
 * end of file menu.c
 */
