/*
 * file menu_network.c - user interface for setting up networks games
 *
 * $Id: menu_network.c,v 1.33 2001/08/16 20:04:00 xblast Exp $
 *
 * Program XBLAST 
 * (C) by Oliver Vogel (e-mail: m.vogel@ndh.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will be entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "menu_network.h"

#include "atom.h"
#include "menu.h"
#include "menu_game.h"
#include "menu_layout.h"
#include "mi_tool.h"
#include "server.h"
#include "str_util.h"

/*
 * local constants
 */ 
#define ANIME_LENGTH 4


/*
 * local variables
 */

/* server configuration */
static char        serverName[128];
static int         serverPort;
static char        serverGame[48];
static CFGGameHost cfgServer;
/* names of all possible players */
static char        *hostName[MAX_HOSTS];         
static const char  *playerName[MAX_HOSTS][NUM_LOCAL_PLAYER];
static XBHostState  hostState[MAX_HOSTS];
static XBHostState  oldHostState[MAX_HOSTS];
static unsigned     hostComplete[MAX_HOSTS];
static int          hostPing[MAX_HOSTS];
/* host buttons */
static MENU_ID  hostItem[MAX_HOSTS];
static MENU_ID  lastFocus = 0;
static unsigned lastHost  = MAX_HOSTS;
/* player graphics */
static XBBool                   playerValid[MAX_HOSTS][NUM_LOCAL_PLAYER];
static CFGPlayerGraphics        playerGfx[MAX_HOSTS][NUM_LOCAL_PLAYER];
static const CFGPlayerGraphics *pPlayerGfx[NUM_LOCAL_PLAYER];
static BMSpriteAnimation        playerAnime[ANIME_LENGTH] = {
    SpriteStopDown, SpriteWinner3,  SpriteWinner2, SpriteWinner, 
};
/* network games */
static const XBNetworkGame *networkGame[NUM_SEARCH_ROWS];
static MENU_ID              networkItem[NUM_SEARCH_ROWS];
static XBAtom               networkAtom;

/*
 *
 */
static void
ClearHosts (void)
{
  int i;

  for (i = 0; i < MAX_HOSTS; i ++) {
    if (NULL != hostName[i]) {
      free (hostName[i]);
      hostName[i] = NULL;
    }
  }
} /* ClearHosts */

/*
 * start network game as client 
 */
static XBBool
ButtonClientStart (void *par)
{
  ClearHosts ();
  /* clean up */
  MenuClear ();
  return XBTrue;
} /* ButtonClientStart */

/*
 * start network game as server
 */
static XBBool
ButtonServerStart (void *par)
{
  int            i, j, k;
  XBPlayerHost   host;
  XBAtom         gameAtom;
  CFGGamePlayers cfgAllPlayers, cfgHostPlayers;
  CFGGameSetup   cfgSetup;

  /* retrieve all player configs */
  j = 0;
  for (i = 0, host = XBPH_Server; i < MAX_HOSTS; i ++, host ++) {
    if (i == 0) {
      gameAtom = atomLocal;
    } else if (hostState[i] == XBHS_In) {
      gameAtom = atomArrayHost0[i];
    } else {
      continue;
    }
    if (! RetrieveGamePlayers (CT_Remote, gameAtom, &cfgHostPlayers) ) {
      continue;
    }
    for (k = 0; k < cfgHostPlayers.num; k ++) {
      cfgAllPlayers.player[j]  = Network_GetPlayer (i, k);
      cfgAllPlayers.control[j] = cfgHostPlayers.control[k];
      cfgAllPlayers.host[j]    = host;
      cfgAllPlayers.team[j]    = XBPT_None;
      j ++;
    }
  }
  cfgAllPlayers.num = j;
  /* and store it */
  StoreGamePlayers (CT_Remote, atomArrayHost0[0], &cfgAllPlayers);
  /* now the other configs */
  if (RetrieveGameSetup (CT_Remote, atomLocal, &cfgSetup)) {
    StoreGameSetup (CT_Remote, atomArrayHost0[0], &cfgSetup);
  }
  /* close port for listening */
  Server_StopListen ();
  /* send data to clients */
  for (i = 1; i < MAX_HOSTS; i ++) {
    switch (hostState[i]) {
    case XBHS_In:
      Server_SendStart (i);
      break;
    default:
      Server_SendDisconnect (i);
      break;
    }
  }
  /* clean up */
  ClearHosts ();
  MenuClear ();
  return XBTrue;
} /* ButtonServerStart */

/*
 * handle network event "player config"
 */
static void
HandlePlayerConfig (unsigned clientID, int player)
{
  XBAtom playerAtom;

  assert (clientID < MAX_HOSTS);
  assert (player >= 0);
  assert (player < NUM_LOCAL_PLAYER);
  /* get player graphics */
  playerAtom                    = Network_GetPlayer (clientID, player);
  playerName[clientID][player]  = GetPlayerName (CT_Remote, playerAtom);
  playerValid[clientID][player] = XBTrue;
  RetrievePlayerGraphics (CT_Remote, playerAtom, COLOR_INVALID, &playerGfx[clientID][player]); 
  /* todo: check completeness */
  hostComplete[clientID] &= ~ (1 << player);
  Dbg_Out ("HandlePlayerConfig: %u %02u\n", clientID, hostComplete[clientID]);
  if (0 == hostComplete[clientID]) {
    hostState[clientID] = XBHS_In;
    MenuSetActive (hostItem[clientID], XBTrue);
  }
} /* HandlePlayerConfig */

/*
 * handle network event "game config"
 */
static void
HandleGameConfig (unsigned clientID)
{
  int i;
  CFGGamePlayers cfgPlayers;

  assert (clientID < MAX_HOSTS);
  /* get host name */
  if (NULL != hostName[clientID]) {
    free (hostName[clientID]);
  }
  hostName[clientID]     = DupString (GetHostName (CT_Remote, atomArrayHost0[clientID]));
  hostState[clientID]    = XBHS_Wait;
  hostComplete[clientID] = 0;
  /* start check for completion */
  if (RetrieveGamePlayers (CT_Remote, atomArrayHost0[clientID], &cfgPlayers) ) {
    for (i = 0; i < cfgPlayers.num && i < NUM_LOCAL_PLAYER; i ++) {
      if (ATOM_INVALID == cfgPlayers.player[i]) {
	break;
      }
      hostComplete[clientID] |= (1 << i);
    }
  }
  Dbg_Out ("HandleGameConfig: %u %02X\n", clientID, hostComplete[clientID]);
  /* resend hoststates for other clients */
  memset (oldHostState, 0, sizeof (oldHostState));
} /* HandleGameConfig */

/*
 * handle network event "disconnected"
 */
static void
HandleShutdown (unsigned clientID)
{
  assert (clientID < MAX_HOSTS);
  /* cleat host entry */
  if (NULL != hostName[clientID]) {
    free (hostName[clientID]);
    hostName[clientID] = NULL;
  }
  hostState[clientID] = XBHS_None;
  hostPing[clientID]  = -1;
  /* clear player data */
  memset (playerValid[clientID], 0, sizeof (playerValid[clientID]));
  memset (playerName[clientID],  0, sizeof (playerName[clientID]));
  /* disable button */
  MenuSetActive (hostItem[clientID], XBFalse);
  /* exit menu if server has disconnected */
  if (0 == clientID) {
    Client_Disconnect ();
    MenuExecFunc (CreateMainMenu, NULL);
  }
} /* HandleShutdown */

/*
 * Handle network event "start game"
 */ 
static void
HandleStartGame (unsigned clientID)
{
  assert (clientID < MAX_HOSTS);

  SetHostType (XBPH_Client1 + clientID - 1);
  MenuExecFunc (ButtonClientStart, NULL);
} /* HandleStartGame */

/*
 * gandle network events "host state"
 */
static void
HandleHostState (unsigned clientID, XBBool isIn)
{
  assert (clientID < MAX_HOSTS);
  if (isIn) {
    hostState[clientID] = XBHS_In;
  } else {
    hostState[clientID] = XBHS_Out;
  }
  Dbg_Out ("server changed host %u state to %s\n", clientID, isIn ? "in" : "out");
} /* HandleHostState */

/*
 * handle network event "ping received"
 */
static void
HandlePingReceived (unsigned clientID, XBBool isServer)
{
  assert (clientID < MAX_HOSTS);
  if (isServer) {
    hostPing[clientID] = Server_GetPingTime (clientID);
  } else {
    hostPing[clientID] = Client_GetPingTime (clientID);
  }
} /* HandlePingReceived */

/*
 * Handle any networks events
 */
static void
HandleNetworkEvents (XBBool isServer)
{
  XBNetworkEvent msg;
  unsigned       clientID;

  while (XBNW_None != (msg = Network_GetEvent (&clientID) ) ) {
    switch (msg) {
      /* player data received */
    case XBNW_RightPlayerConfig: HandlePlayerConfig (clientID, 0); 	  break;
    case XBNW_LeftPlayerConfig:  HandlePlayerConfig (clientID, 1); 	  break;
    case XBNW_Joy1PlayerConfig:  HandlePlayerConfig (clientID, 2); 	  break;
    case XBNW_Joy2PlayerConfig:  HandlePlayerConfig (clientID, 3); 	  break;
      /* game config received */       
    case XBNW_GameConfig:        HandleGameConfig (clientID);      	  break;
      /* one host has disconnected */	    
    case XBNW_Disconnected:      HandleShutdown (clientID);        	  break;
      /* error in connection to host */	      
    case XBNW_Error:             HandleShutdown (clientID);         	  break;
      /* ping received from client */
    case XBNW_PingReceived:      HandlePingReceived (clientID, isServer); break;
      /* server wants to start game */
    case XBNW_StartGame:         HandleStartGame (clientID);              break;
      /* server has changed host state */
    case XBNW_HostIsIn:          
      if (! isServer) {
	HandleHostState (clientID, XBTrue);
      }
      break;
    case XBNW_HostIsOut:          
      if (! isServer) {
	HandleHostState (clientID, XBFalse);
      }
      break;
      /* anything else */
    default:
      break;
    }
  }
} /* HandleNetworkEvents */ 

/*
 *
 */
static void
UpdateFocus (void)
{
  MENU_ID  newFocus;
  unsigned newHost;
  int      player;

  newFocus = MenuGetFocus ();
  if (lastFocus != newFocus) {
    if (0 != newFocus) {
      /* which host */
      for (newHost = 0; newHost < MAX_HOSTS; newHost ++) {
	if (hostItem[newHost] == newFocus) {
	  if (newHost != lastHost) {
	    for (player = 0; player < NUM_LOCAL_PLAYER; player ++) {
	      if (playerValid[newHost][player]) {
		pPlayerGfx[player] = &playerGfx[newHost][player];
	      } else {
		pPlayerGfx[player] = NULL;
	      }
	    }
	    lastHost = newHost;
	  }
	  break;
	}
      }
    }
    lastFocus = newFocus;
  }
} /* UpdateFocus */

/*
 *
 */
static void
CheckHostStates (void)
{
  unsigned i;

  for (i = 0; i < MAX_HOSTS; i ++) {
    if (oldHostState[i] != hostState[i]) {
      /* inform clients about state chnage */
      switch (hostState[i]) {
      case XBHS_In:  Server_SendHostState (i, XBTrue);  break;
      case XBHS_Out: Server_SendHostState (i, XBFalse); break;
      default:       break;
      }
      oldHostState[i] = hostState[i];
    }
  }
} /* CheckHostStates (void) */

/*
 *
 */
static void
PollNetwork (void *par)
{
  XBBool isServer;
  XBAtom *atom = par;
  assert (atom != NULL);

  isServer = (*atom == atomServer);
  /* check network events */
  HandleNetworkEvents (isServer);
  /* show players according to selected host */
  UpdateFocus ();
  /* check host states if server */
  if (isServer) {
    CheckHostStates ();
  }
} /* PollNetwork */

/*
 * 
 */
static XBBool
ButtonServerDisconnect (void *par)
{
  int i;
  /* close port for listening */
  Server_StopListen ();
  /* finish communication with clients */
  Server_SendDisconnectAll ();
  /* clean up */
  for (i = 1; i < MAX_HOSTS; i ++) {
    if (NULL != hostName[i]) {
      free (hostName[i]);
    }
  }
  /* back to connection menu */
  return CreateServerMenu (par);
} /* ButtonServerDisconnect */

/*
 * 
 */
static XBBool
ButtonServerKickOut (void *par)
{
  int i;

  /* disconnect unwanted clients  */
  for (i = 1; i < MAX_HOSTS; i ++) {
    if (hostState[i] == XBHS_Out) {
      /* disconnect this client */
      Server_SendDisconnect (i);
      /* simluate disconnect from client */
      Server_ReceiveDisconnect (i);
    }
  }
  /* back to connection menu */
  return XBFalse;
} /* ButtonServerDisconnect */

/*
 * local initializations
 */
static void
LocalInit (void)
{
  memset (hostName,     0, sizeof (hostName));
  memset (hostState,    0, sizeof (hostState));
  memset (oldHostState, 0, sizeof (oldHostState));
  memset (hostComplete, 0, sizeof (hostComplete));
  memset (playerName,   0, sizeof (playerName));
  memset (playerGfx,    0, sizeof (playerGfx));
  memset (playerValid,  0, sizeof (playerValid));
  memset (hostPing,    -1, sizeof (hostPing));
  lastFocus   = 0;
  lastHost    = MAX_HOSTS;
} /* LocalInit */

/*
 * create player graphics
 */
static void
CreatePlayerItems (void)
{
  int i;

  for (i = 0; i < NUM_LOCAL_PLAYER; i ++) {
    MenuAddPlayer (PLAYER_LEFT (i, NUM_LOCAL_PLAYER), PLAYER_TOP, PLAYER_WIDTH, i, pPlayerGfx + i, -ANIME_LENGTH, playerAnime);
  }
} /* CreatePlayerItems */

/*
 * create host items
 */
static void 
CreateHostItems (XBBool server)
{
  int x, y;
  int player;
  int client = 0;
  int r      = MENU_TOP;
  for (y = 0; y < 3; y ++) {
    int c = 5 * CELL_W / 2;
    for (x = 0; x < 2; x ++) {
      /* host */
      if (0 == client) {
	hostItem[client] = MenuAddServer (c, r, 4*CELL_W, (const char **) hostName + client);
      } else {
	if (server) {
	  hostItem[client] = MenuAddClient (c, r, 4*CELL_W, (const char **) hostName + client, hostState + client, hostPing + client);
	} else {
	  hostItem[client] = MenuAddPeer (c, r, 4*CELL_W, (const char **) hostName + client, hostState + client, hostPing + client);
	}
	MenuSetActive (hostItem[client], XBFalse);
      }
      /* players */
      player = 0;
      MenuAddTag (c - 1*CELL_W, r + 2*CELL_H/2, 2*CELL_W, &playerName[client][player ++]);
      MenuAddTag (c - 1*CELL_W, r + 3*CELL_H/2, 2*CELL_W, &playerName[client][player ++]);
      MenuAddTag (c + 1*CELL_W, r + 2*CELL_H/2, 2*CELL_W, &playerName[client][player ++]);
      MenuAddTag (c + 1*CELL_W, r + 3*CELL_H/2, 2*CELL_W, &playerName[client][player ++]);
      MenuAddTag (c + 3*CELL_W, r + 2*CELL_H/2, 2*CELL_W, &playerName[client][player ++]);
      MenuAddTag (c + 3*CELL_W, r + 3*CELL_H/2, 2*CELL_W, &playerName[client][player ++]);
      client ++;
      c += 6*CELL_W;
    }
    r += 2*CELL_H;
  }  
} /* CreateHostItems */

/*
 *
 */
static XBBool
CreateServerWaitMenu (void *par) 
{
  XBAtom *atom = par;
  int player;
  CFGGamePlayers cfgGame;

  assert (atom != NULL);
  /* inits */
  LocalInit ();
  hostName[0] = DupString ("localhost");
  /* get local player data */
  if (RetrieveGamePlayers (CT_Local, atomServer, &cfgGame) ) {
    for (player = 0; player < cfgGame.num; player ++) {
      playerValid[0][player] = XBTrue;
      RetrievePlayerGraphics (CT_Local, cfgGame.player[player], COLOR_INVALID, &playerGfx[0][player]); 
      playerName[0][player] = GetPlayerName (CT_Local, cfgGame.player[player]);
    }
  }
  /* build menu */
  MenuClear ();
  /* poll function */
  MenuAddCyclic (PollNetwork, par);
  /* players */
  CreatePlayerItems ();
  /* Title */
  MenuAddLabel (TITLE_LEFT, TITLE_TOP, TITLE_WIDTH, "Waiting for Clients");
  /* hosts */
  CreateHostItems (XBTrue);
  /* Buttons */
  MenuSetAbort   (MenuAddHButton ( 3 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Disconnect", ButtonServerDisconnect, par) );
  (void)          MenuAddHButton (11 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Kick out",   ButtonServerKickOut,    par);
  MenuSetDefault (MenuAddHButton (19 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Start",      ButtonServerStart,      NULL) );
  /* --- */
  MenuSetLinks ();
  /* that's all*/
  return XBFalse;
} /* CreateServerWaitMenu */

/*
 * 
 */
static XBBool
ButtonClientDisconnect (void *par)
{
  /* finish communication to server */
  Client_Disconnect ();
  /* back to connection menu */
  return CreateJoinNetGameMenu (par);
} /* ButtonDisconnect */

/*
 *
 */
static XBBool
CreateClientWaitMenu (void *par) 
{
  XBAtom *atom = par;
  
  assert (atom != NULL);
  /* needed inits */
  LocalInit ();
  /* build menu */
  MenuClear ();
  /* poll function */
  MenuAddCyclic (PollNetwork, par);
  /* players */
  CreatePlayerItems ();
  /* Title */
  MenuAddLabel (TITLE_LEFT, TITLE_TOP, TITLE_WIDTH, "Game Setup by Server");
  /* hosts */
  CreateHostItems (XBFalse);
  /* Buttons */
  MenuSetAbort (MenuAddHButton (9*CELL_W/2, MENU_BOTTOM, 6*CELL_W, "Disconnect", ButtonClientDisconnect, par) );
  /* --- */
  MenuSetLinks ();
  /* that's all*/
  return XBFalse;
} /* CreateClientWaitMenu */

/*
 *
 */
XBBool
ButtonServerListen (void *par)
{
  XBAtom *atom = par;

  assert (atom != NULL);
  /* set server connection */
  cfgServer.name = NULL;
  cfgServer.port = serverPort;
  cfgServer.game = serverGame;
  /* store in database */
  StoreGameHost (CT_Local, *atom, &cfgServer);
  /* get game setup */
  /* now try connection */
  if (! Server_StartListen (&cfgServer) ) {
    /* connection failed */
    GUI_ErrorMessage ("Failed to listen on port %hu", serverPort);
    return CreateServerMenu (par);
  }
  return CreateServerWaitMenu (par);
} /* ButtonServerListen */

/*
 * setup xblast server
 */
XBBool
CreateServerMenu (void *par)
{
  XBAtom *atom = par;

  assert (atom != NULL);
  /* get old server from database */
  if (! RetrieveGameHost (CT_Local, *atom, &cfgServer)) {
    serverPort    = 16168;
    serverGame[0] = 0;
  } else {
    serverPort = cfgServer.port;
    strncpy (serverGame, cfgServer.game, sizeof (serverGame));
    serverGame[sizeof (serverGame) - 1] = 0;
  }
  /* build menu */
  MenuClear ();
  /* Title */
  MenuAddLabel (TITLE_LEFT, TITLE_TOP, TITLE_WIDTH, "Setup Server");
  /* parameters */
  MenuAddInteger   (DLG_LEFT, MENU_ROW (0), DLG_WIDTH, "TCP-Port:",               2*CELL_W, &serverPort, 4096, 65535);
  MenuAddComboBool (DLG_LEFT, MENU_ROW (1), DLG_WIDTH, "Use fixed UDP-ports",     2*CELL_W, &cfgServer.fixedUdpPort);
  MenuAddComboBool (DLG_LEFT, MENU_ROW (2), DLG_WIDTH, "Allow clients using NAT", 2*CELL_W, &cfgServer.allowNat);
  MenuAddComboBool (DLG_LEFT, MENU_ROW (3), DLG_WIDTH, "Visible in LAN:",         2*CELL_W, &cfgServer.browseLan);
  MenuAddString    (DLG_LEFT, MENU_ROW (4), DLG_WIDTH, "Name:",                   4*CELL_W, serverGame, sizeof (serverGame));
  /* Buttons */
  MenuSetAbort   (MenuAddHButton ( 5 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Back",     CreateStartNetGameMenu, par) );
  MenuSetDefault (MenuAddHButton (17 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Continue", ButtonServerListen,     par) );
  /* --- */
  MenuSetLinks ();
  /* that's all*/
  return XBFalse;
} /* CreateServerMenu */

/*
 * Connect to server
 */
static XBBool
ButtonClientConnect (void *par)
{
  XBAtom *atom = par;

  assert (atom != NULL);
  /* get server connection */
  cfgServer.name = serverName;
  cfgServer.port = serverPort;
  /* store in database */
  StoreGameHost (CT_Local, *atom, &cfgServer);
  /* now try connection */
  if (! Client_Connect (&cfgServer) ) {
    /* connection failed */
    GUI_ErrorMessage ("Connection to %s:%hu failed", serverName, serverPort);
    return CreateClientMenu (par);
  }
  return CreateClientWaitMenu (par);
} /* ButtonClientConnect */

/*
 * setup server connection
 */
XBBool 
CreateClientMenu (void *par)
{
  XBAtom *atom = par;

  assert (atom != NULL);
  /* get old server from database */
  if (! RetrieveGameHost (CT_Local, *atom, &cfgServer)) {
    serverName[0] = 0;
    serverPort    = 16168;
  } else {
    strncpy (serverName, cfgServer.name, sizeof (serverName));
    serverName[sizeof (hostName) - 1] = 0;
    serverPort = cfgServer.port;
  }
  /* build menu */
  MenuClear ();
  /* Title */
  MenuAddLabel (TITLE_LEFT, TITLE_TOP, TITLE_WIDTH, "Connect to Server");
  /* parameters */
  MenuAddString  (DLG_LEFT, MENU_ROW (0), DLG_WIDTH, "Hostname:", 4*CELL_W, serverName, sizeof (serverName));
  MenuAddInteger (DLG_LEFT, MENU_ROW (1), DLG_WIDTH, "TCP-Port:", 4*CELL_W, &serverPort, 4096, 65535);
  /* Buttons */
  MenuSetAbort   (MenuAddHButton ( 5 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Back",    CreateJoinNetGameMenu, par) );
  MenuSetDefault (MenuAddHButton (17 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Connect", ButtonClientConnect,   par) );
  /* --- */
  MenuSetLinks ();
  /* that's all*/
  return XBFalse;
} /* CreateClientMenu */

/*
 * poll incoming network games
 */
static void
PollNetworkGame (void *par)
{
  unsigned       gameID;
  XBNetworkEvent msg;

  while (XBNW_None != (msg = Network_GetEvent (&gameID) ) ) {
    if (msg == XBNW_NetworkGame) {
      const XBNetworkGame *game = Client_NextNetworkGame ();
      if (gameID < NUM_SEARCH_ROWS) {
	networkGame[gameID] = game;
	MenuSetActive (networkItem [gameID], (NULL != game));
      } 
    }
  }
} /* PollNetworkGame */

/*
 * 
 */
static XBBool 
ButtonExitSearchLan (void *par)
{
  /* stop network search */
  Client_StopQuery ();
  /* back to client menu */
  return CreateJoinNetGameMenu (par);
} /* ButtonExitSearchLan */

/*
 *
 */
static XBBool
ButtonRefreshSearchLan (void *par)
{
  unsigned i;

  /* delete old games */
  memset (networkGame, 0, sizeof (networkGame));
  for (i = 0; i < NUM_SEARCH_ROWS; i ++) {
    MenuSetActive (networkItem[i], XBFalse);
  }
  /* starts new query */
  Client_RestartQuery ();
  return XBFalse;
} /* ButtonRefreshSearchLan */

/*
 *
 */
static XBBool
ButtonStartLanGame (void *par)
{
  const XBNetworkGame **ptr = par;
  XBBool connect;

  assert (NULL != ptr);
  if (NULL == *ptr) {
    return XBFalse;
  }
  /* get server connection */
  cfgServer.name = (*ptr)->host;
  cfgServer.port = (*ptr)->port;
  /* store in database */
  StoreGameHost (CT_Local, networkAtom, &cfgServer);
  /* now try connection */
  connect = Client_Connect (&cfgServer);
  /* clean up */
  Client_StopQuery ();
  memset (networkGame, 0, sizeof (networkGame));
  if (! connect) {
    /* connection failed */
    GUI_ErrorMessage ("Connection to %s:%hu failed", serverName, serverPort);
    return CreateSearchLanMenu (&networkAtom);
  }
  return CreateClientWaitMenu (&networkAtom);

} /* ButtonStartNetworkGame */

/*
 * search LAN for network games
 */
XBBool
CreateSearchLanMenu (void *par)
{
  XBAtom *atom = par;
  int i;

  assert (NULL != atom);
  networkAtom = *atom;
  /* setup */
  memset (networkGame, 0, sizeof (networkGame));
  (void) RetrieveGameHost (CT_Local, networkAtom, &cfgServer);
  /* build menu */
  MenuClear ();
  /* Title */
  MenuAddLabel (TITLE_LEFT, TITLE_TOP, TITLE_WIDTH, "Search LAN for Games");
  /* list of games */
  MenuSetActive (MenuAddGameHeader (SEARCH_LEFT, SEARCH_TOP, SEARCH_WIDTH), XBFalse);
  for (i = 0; i < NUM_SEARCH_ROWS; i ++) {
    networkItem[i] = MenuAddGameEntry (SEARCH_LEFT, SEARCH_ROW(i), SEARCH_WIDTH, networkGame + i, ButtonStartLanGame);
    MenuSetActive (networkItem[i], XBFalse);
  }
  /* Buttons */
  MenuSetAbort   (MenuAddHButton ( 5 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Back",    ButtonExitSearchLan,    par) );
  MenuSetDefault (MenuAddHButton (17 * CELL_W/2, MENU_BOTTOM, 4*CELL_W, "Refresh", ButtonRefreshSearchLan, par) );
  /* add polling routine */
  MenuAddCyclic (PollNetworkGame, NULL);
  /* query for local games */
  Client_StartQuery ();
  /* that's all*/
  return XBFalse;
} /* CreateSearchLanMenu */

/*
 * end of file menu_network.c
 */
