/*
 * file status.c - displays status bar at window bottom
 *
 * Program XBLAST 
 * (C) by Oliver Vogel (e-mail: m.vogel@ndh.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will be entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include "status.h"

#include "map.h"
#include "geom.h"

#include "gui.h"

/*
 * local constants
 */
#define TEXT_ATTR_NORMAL (FF_Medium|FF_White|FF_Boxed)
#define TEXT_ATTR_NUMBER (FF_Large|FF_White|FF_Boxed)

#define MAX_MSG           8
#define MESSAGE_TIME      TIME_STEP

/*
 * local variables
 */
static int numPlayer;

static int gameNext, timeLed;
static int faceUsed[MAX_PLAYER];
static int numUsed[MAX_PLAYER];
static const int facePos[MAX_PLAYER] = {
  0, 2, 4, 14, 16, 18,
};
static const int numPos[MAX_PLAYER] = {
  1, 3, 5, 15, 17, 19,
};

/* variables for message queue */
static const char *defaultMessage;
static int         msgRestore;
static int         numMsg = 0;
static int         firstMsg;
static const char *msgQueue[MAX_MSG];
static const int   msgDelay[MAX_MSG+1] = {
  -1, 50, 40, 20, 12, 7, 7, 7, 7
};

/* status board text box */
static const int textLeft  = 6;
static const int textRight = 13;
static BMRectangle  statusBox = {
  77*STAT_WIDTH/12, 79*STAT_HEIGHT/6,
  87*STAT_WIDTH/12,  2*STAT_HEIGHT/3,
};

/* status board number boxes */
static BMRectangle statusNumberBox [STAT_W] = {
  {   1*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  {  13*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  {  25*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  {  37*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  {  49*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  {  61*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  {  73*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  {  85*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  {  97*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 109*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 121*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 133*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 145*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 157*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 169*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 181*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 193*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 205*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 217*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
  { 229*STAT_WIDTH/12, 79*STAT_HEIGHT/6, 8*STAT_WIDTH/12-1, 4*STAT_HEIGHT/6 },
};

/*
 *
 */
void
ClearStatusBar (int tile1, int tile2)
{
  int x;

  for (x = 0; x < MAZE_W; x ++) {
    GUI_DrawBlock (x, MAZE_H,   tile1);
    GUI_DrawBlock (x, MAZE_H+1, tile2);
  }
  MarkMazeRect (0, PIXH, PIXW, SCOREH);
} /* ClearStatusBar */

/*
 *
 */
static void
SetScoreBlock (int x, int block)
{
  GUI_DrawScoreBlock (x, block);
  MarkMazeTile (x, MAZE_H);
} /* SetScoreBlock */

/*
 *
 */
static void 
SetStatusText (const char *msg)
{
  GUI_DrawSimpleTextbox (msg, TEXT_ATTR_NORMAL, &statusBox);
  MarkMaze (textLeft, MAZE_H, textRight, MAZE_H);
} /* SetStatusText */

/*
 *
 */
static void
SetStatusNumber (int pos, int value)
{
  char numString[2];

  numString[0] = '0' + value;
  numString[1] = '\0';
  GUI_DrawSimpleTextbox (numString, TEXT_ATTR_NUMBER, statusNumberBox + pos);
  MarkMazeTile (pos, MAZE_H);
} /* SetStatusNumber */

/*
 *
 */
static void
SetTimeLed (int x, int val)
{
  GUI_DrawTimeLed (x, val);
  MarkMazeTile (x/3, MAZE_H+1);
} /* SetTimeLed */

/*
 *
 */
void
InitStatusBar (int numPlayers) 
{
  int p;

  /* set player and display number */
  numPlayer = numPlayers;

  /* clear all */
  for (p = 0; p < STAT_W; p ++) {
    SetScoreBlock (p, SBVoid);
  }

  /* player boxes */
  for (p = 0; p < numPlayer; p ++) {
    faceUsed[p] = SBPlayer + p;
    numUsed[p]  = 0;
#if 0
    SetScoreBlock   (facePos[p], p);
#endif
    SetScoreBlock   (facePos[p], faceUsed[p]);
    SetScoreBlock   (numPos[p],  SBTextRight);
    SetStatusNumber (numPos[p],  numUsed[p]);
  }
  /* text box */
  SetScoreBlock (textLeft,  SBTextLeft);
  for (p = textLeft + 1; p < textRight; p ++) {
    SetScoreBlock (p, SBTextMid);
  }
  SetScoreBlock (textRight, SBTextRight);
  /* set queue to default message */
  defaultMessage = "";
  numMsg         = 0;
  SetStatusText (defaultMessage);

  /* set leds */
  for (p = 0; p < (4*MAZE_W); p ++) {
    SetTimeLed (p, 1);
  }
  /* set timers for leds */
  gameNext = TIME_STEP;
  timeLed  = MAZE_W*4;
} /* InitStatusBar */

/* 
 *
 */
void
ResetStatusBar (const BMPlayer *ps, const char *msg, XBBool flag)
{
  int p;

  /* player boxes */
  for (p = 0; p < numPlayer; p ++) {
    faceUsed[p] = SBPlayer + p;
    numUsed[p]  = flag ? ps[p].lives : ps[p].victories;
    SetScoreBlock   (facePos[p], faceUsed[p]);
    SetStatusNumber (numPos[p],  numUsed[p]);
  }
  /* default message */
  defaultMessage = msg;
  numMsg         = 0;
  SetStatusText (msg);
  /* set leds */
  for (p = 0; p < 4*MAZE_W; p ++) {
    SetTimeLed (p, 1);
  }
  /* set timers for leds */
  gameNext = TIME_STEP;
  timeLed  = MAZE_W*4;
  /* marl all filled to be redrawed */
  MarkMaze (0, MAZE_H, MAZE_W, MAZE_H+1);
} /* ResetStatusBar */

/*
 *
 */
void
UpdateStatusBar (const BMPlayer *ps, int game_time)
{
  int p;
  int newFace;
  
  for (p = 0; p < numPlayer; p ++) {
    /* determine current face for player */
    if (ps[p].dying || ps[p].lives == 0) {
      newFace = SBDead   + p;
    } else if (ps[p].illness != ps[p].health) {
      newFace = SBSick   + p;
    } else {
      newFace = SBPlayer + p;
    }
    if (faceUsed[p] != newFace) {
      faceUsed[p] = newFace;
      SetScoreBlock (facePos[p], faceUsed[p]);
    }
    /* check player lives */
    if (numUsed[p] != ps[p].lives) {
      numUsed[p] = ps[p].lives;
      SetStatusNumber (numPos[p], numUsed[p]);
    }
  }
  /* check leds */
  if (game_time > gameNext) {
    gameNext += TIME_STEP;
    timeLed  --;
    SetTimeLed (timeLed, 0);
  }
  /* check message */
  /* if any message is in the queue */
  if (numMsg) {
    msgRestore ++;
    /* check if message is outdated */
    if (msgRestore >= msgDelay[numMsg]) {
      /* return to default message if last queued one */
      if (-- numMsg) {
	/* set next message */
	firstMsg = (firstMsg+1) % MAX_MSG;
	SetStatusText (msgQueue[firstMsg]);
	msgRestore = 0;
      } else {
	/* return to default message if last queued one */
	SetStatusText (defaultMessage);
      }
    }
  }
} /* UpdateStatusBar */

/*
 * 
 */
void
SetMessage (const char *msg, XBBool perm)
{
  int m;

  if (NULL != msg) {
    if (perm) {
      SetStatusText (msg);
      numMsg = 0;
    } else {
      /* if no other message exists set to start of queue */
      if (0 == numMsg) {
	m = firstMsg = 0;
	msgRestore = 0;
	SetStatusText (msg);
      } else {
	m = (firstMsg + numMsg) % MAX_MSG;
      }
      msgQueue[m] = msg;
      numMsg ++;
      /* check for queue overflow */
      if (numMsg > MAX_MSG) {
	/* remove first message */
	numMsg --;
	firstMsg = (firstMsg+1) % MAX_MSG;
      }
    }
  }
} /* SetMessage */

/*
 * 
 */
void
ResetMessage (void)
{
  if (numMsg > 0) {
    SetStatusText (msgQueue[firstMsg]);
  } else {
    SetStatusText (defaultMessage);
  }
} /* ResetMessage */

/*
 * end of file status.c
 */
